/*
 * cdata.cpp
 * Copyright (C) Raul Massano Brás 2012 <rbras@iseg.utl.pt>
 *
 * MulTyLink is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MulTyLink is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "cdata.h"
#include "stringutils.h"
#include <fstream>
#include <QMessageBox>
#include <QCoreApplication>
#include <algorithm>

using namespace std;


int binSearchSites(const vector<site_info *>& sites, int val)
{
  //Because we keep sites sorted by id we do a binary search here
  int left=0,right=(int)sites.size()-1;
  if(right<0) return -1;
  int p=(right+left)/2;
  for(;;) {
    if(right-left<8) { //do a linear search
      for(int i=left;i<=right;i++) if(sites[i]->id==val) return i;
      return -1;
    }
    if(sites[p]->id==val) return p;
    else if(sites[p]->id<val) left=p+1;
    else right=p-1;
    p=(right+left)/2;
  }

}

CData::CData(MainWindow *mf)
{
    mainFrm=mf;
    m=n=0;
    mat=NULL;
    names=NULL;
    pa_ids=NULL;
    adjDist=NULL;
    numTypes=0;
    numSites=0;
    zoom=1;
    maxId=0;
    maxFriction=-1e+100;
}
CData::~CData()
{


    if(mat) {
        for(size_t i=0;i<m;i++) {
            for(size_t j=0;j<n;j++) {
                if(mat[i][j]) {
                    if(mat[i][j]->vd) {
                        if(mat[i][j]->vd->barriers) delete [] mat[i][j]->vd->barriers;
                        if(mat[i][j]->vd->termId) delete [] mat[i][j]->vd->termId;
                        delete mat[i][j]->vd;
                    }
                    delete mat[i][j];
                }
            }
            delete [] mat[i];
        }
        delete [] mat;
    }
    if(names) delete [] names;
    if(pa_ids) delete [] pa_ids;
    for(size_t i=0;i<fake_sites.size();i++) {
        delete fake_sites[i]->vd;
        delete fake_sites[i]->vd->barriers;
        delete fake_sites[i];
    }
    if(adjDist) delete [] adjDist;
}

bool CData::readFile(const QString& file)
{

    //Read file with site info and adjacencies and build the graph with all vertices
    //and adjcencies

    ifstream in(file.toStdString().c_str());
    if(!in) {
        QMessageBox::critical(NULL,tr("error"),tr("Could not open %1").arg(file));
        return false;
    }

    //File uses "." for decimal separator and "," for field separator
    //so be carefull with numeric format

    setlocale(LC_NUMERIC,"en_US");

    mainFrm->logMsg(tr("Reading file %1").arg(file));

    vector<string> tk;
    size_t line=0;

    //Number of sites, number of terminal types

    if(!readLine(in,tk,",",line)) {
        QMessageBox::critical(NULL,tr("error"),tr("Unexpected end of file at line %1").arg(line));
        in.close();
        return false;
    }
    if(tk.size()!=2) {
        QMessageBox::critical(NULL,tr("error"),tr("Error reading the number of sites and types file %1").arg(file));
        in.close();
        return false;
    }

    numSites=atoi(tk[0].c_str());
    numTypes=atoi(tk[1].c_str());  //number of terminal types

    mainFrm->logMsg(tr("Number of sites: %1").arg(numSites));
    mainFrm->logMsg(tr("Number of types: %1").arg(numTypes));

    QCoreApplication::processEvents();


    //Read the third line
    //Min X coord, Max X, Min Y, Max Y, step, zoom factor

    if(!readLine(in,tk,",",line)) {
        QMessageBox::critical(NULL,tr("error"),tr("Unexpected end of file at line %1").arg(line));
        in.close();
        return false;
    }

    if(tk.size()!=6) {
        QMessageBox::critical(NULL,tr("error"),tr("Error reading the coordinates at line %1").arg(line));
        in.close();
        return false;
    }
    const double minX=atof(tk[0].c_str());
    const double maxX=atof(tk[1].c_str());
    const double minY=atof(tk[2].c_str());
    const double maxY=atof(tk[3].c_str());
    const double step=atof(tk[4].c_str());
    zoom=atoi(tk[5].c_str());
    if(zoom==0) zoom=1;
    if(zoom>16) {
        mainFrm->logMsg(tr("Maximum zoom is 16: %1 changed to 16").arg(zoom));
        zoom=16;
    }

    const size_t maxCoordX=(size_t)((maxX-minX)/step+0.5);
    const size_t maxCoordY=(size_t)((maxY-minY)/step+0.5);

    //Read the adjacency distances

    if(!readLine(in,tk,",",line)) {
        QMessageBox::critical(NULL,tr("error"),tr("Unexpected end of file at line %1").arg(line));
        in.close();
        return false;
    }

    if(tk.size()!=numTypes) {
        QMessageBox::critical(NULL,tr("error"),tr("Error reading adjacencies distances at line %1. You must specify %2 distances").arg(line).arg(numTypes));
        in.close();
        return false;
    }
    adjDist=new double [numTypes];

    for(size_t i=0;i<numTypes;++i) {
        adjDist[i]=atof(tk[i].c_str());
        if(adjDist[i]<1) {
            QMessageBox::critical(NULL,tr("error"),tr("Distance must be greater or equal to 1"));
            in.close();
            return false;
        }
        mainFrm->logMsg(tr("Adjacency distance for type %1: %2").arg(i+1).arg(adjDist[i]));
    }
    QCoreApplication::processEvents();

    //Read the name of the PA types
    if(!readLine(in,tk,",",line)) {
        QMessageBox::critical(NULL,tr("error"),tr("Unexpected end of file at line %1").arg(line));
        in.close();
        return false;
    }
    if(tk.size()!=numTypes) {
        QMessageBox::critical(NULL,tr("error"),tr("Error reading the PA types names at line %1").arg(line));
        in.close();
        return false;
    }

    names=new string [numTypes];
    for(size_t i=0;i<numTypes;i++) {
        names[i]=tk[i];
        Trim(names[i]);
    }


    //Now read the numSites sites


    //Create a temporary data structure to store the id of the site
    //We start by storing the site info in a grid. Later this information
    //will be used to calculate adjacencies and build the graph

    m=maxCoordX+1;
    n=maxCoordY+1;

    mat=new site_info **[m];
    for(size_t i=0;i<m;i++) mat[i]=new site_info *[n];
    for(size_t i=0;i<m;i++)
        for(size_t j=0;j<n;j++) mat[i][j]=NULL;

    double posx,posy;
    size_t xcoord,ycoord;

    size_t numRead=0;

    maxId=-1;

    sites.reserve(numSites);
    vector<site_info *> terms[numTypes];

    while(readLine(in,tk,",",line)) {

        if(line%50000==0) mainFrm->logMsg(tr("Reading line %1").arg(line));
        QCoreApplication::processEvents();

        if(tk.size()<7) {
            QMessageBox::critical(NULL,tr("error"),tr("Error in the number of elements at line %1").arg(line));
            in.close();
            return false;
        }

        //Read the id
        int id=atoi(tk[0].c_str());
        if(id<=maxId) {
            QMessageBox::critical(NULL,tr("error"),tr("File must be sorted by site id and ids must be unique. At line %1").arg(line));
            in.close();
            return false;
        }
        maxId=id;

        //Read the coordinates

        posx=atof(tk[1].c_str());
        posy=atof(tk[2].c_str());

        if(posx<minX || posx>maxX || posy<minY || posy>maxY) {
            QMessageBox::critical(NULL,tr("error"),
                                  tr("Error in coordinates at line %1. Please verify the max and min values in file.").arg(line));
            in.close();
            return false;
        }

        xcoord=(size_t)((posx-minX)/step+0.5);
        ycoord=(size_t)((maxY-posy)/step+0.5);

        if( xcoord>maxCoordX || ycoord>maxCoordY ) {
            QMessageBox::critical(NULL,tr("error"),tr("Error converting coordinates at line %1").arg(line));
            in.close();
            return false;
        }


        if(mat[xcoord][ycoord]) { //Site was alread read
            QMessageBox::critical(NULL,tr("error"),tr("Site with id %1 at (%2,%3) overlaps site with id %4").arg(id).
                                  arg(posx).arg(posy).arg(mat[xcoord][ycoord]->id));
            in.close();
            return false;
        }

        //Cost and friction
        double cost=atof(tk[3].c_str());
        double friction=atof(tk[4].c_str());
        if(friction>maxFriction) maxFriction=friction;

        //Number of types the cell belongs to

        int nTypesOfCell=atoi(tk[5].c_str());
        if((int)tk.size()<7+nTypesOfCell) {
            QMessageBox::critical(NULL,tr("error"),tr("Error in the number of elements at line %1").arg(line));
            in.close();
            return false;
        }


        if( (nTypesOfCell<0) || (nTypesOfCell > (int)numTypes)) {
            QMessageBox::critical(NULL,tr("error"),
                                  tr("The number of types for each cell must be between 1 and %1. Found %2 at line %3.").
                                  arg(numTypes).arg(nTypesOfCell).arg(line));
            in.close();
            return false;
        }

        //Types in the cell
        vector<int> cellTypes;
        if(!nTypesOfCell) cellTypes.push_back(0);
        else {
            for(int i=0;i<nTypesOfCell;i++) {
                int cType=atoi(tk[6+i].c_str());
                if(cType<1 || cType>(int)numTypes) {
                    QMessageBox::critical(NULL,tr("error"),
                                          tr("Types for each cell must be between 1 and %1. Found %2 at line %3.").
                                          arg(numTypes).arg(cType).arg(line));
                    in.close();
                    return false;
                }
                cellTypes.push_back(cType);
            }
        }
        sort(cellTypes.begin(),cellTypes.end());

        //Barriers

        bool *B=new bool [numTypes];
        for(size_t i=0;i<numTypes;i++) B[i]=false;

        int nBarriers=atoi(tk[6+nTypesOfCell].c_str());
        if((int)tk.size()!=7+nTypesOfCell+nBarriers) {
            QMessageBox::critical(NULL,tr("error"),tr("Error in the number of elements at line %1").arg(line));
            in.close();
            delete [] B;
            return false;
        }


        for(int i=0;i<nBarriers;i++) {
            int b=atoi(tk[7+nTypesOfCell+i].c_str());
            if(b<1 || b>(int)numTypes) {
                QMessageBox::critical(NULL,tr("error"),
                                      tr("Types for each cell must be between 1 and %1. Found %2 at line %3.").
                                      arg(numTypes).arg(b).arg(line));
                in.close();
                return false;
            }
            //It can't be a barrier to itself
            if(nTypesOfCell) {
                if(binary_search(cellTypes.begin(),cellTypes.end(),b)) {
                    QMessageBox::critical(NULL,tr("error"),
                                          tr("Cells can't be barriers to their own type. At line %1.").arg(line));
                    in.close();
                    delete [] B;
                    return false;
                }
            }
            B[b-1]=true;
        }


        numRead++;

        point p(xcoord,ycoord);
        vertex_data *ud=new vertex_data;

        int type_id;
        if(nTypesOfCell<=1) type_id=cellTypes[0];
        else type_id=numTypes+1;

        ud->cellTypes=cellTypes;

        ud->grid_coords=p;
        ud->siteType=type_id;
        ud->barriers=B;
        //If the site is a protected area its cost is 0 and is allways usable.
        if(type_id>0) {cost=0; friction=0;}
        ud->cost=cost;
        ud->friction=friction;
        ud->xcoord=posx;
        ud->ycoord=posy;
        ud->termId=NULL; //Terminal ids will be assigned later
        if(nTypesOfCell) ud->termId=new int [numTypes];
        mat[xcoord][ycoord]=new site_info;
        mat[xcoord][ycoord]->id=id;
        mat[xcoord][ycoord]->vd=ud;

        vdMap[id]=ud;

        sites.push_back(mat[xcoord][ycoord]); //sites is ordered by id

        for(int i=0;i<nTypesOfCell;i++) {
            terms[cellTypes[i]-1].push_back(mat[xcoord][ycoord]);
        }

    }
    in.close();


    if(numRead!=numSites) {
        QMessageBox::critical(NULL,tr("error"),tr("Expected %1 sites, read %2").arg(numSites).arg(numRead));
        return false;
    }

    //Terminals of the same type that are connected are a uniq terminal

    size_t contador=1;
    pa_ids=new vector<int> [numTypes];
    for(size_t k=0;k<numTypes;k++) {
        //Build a graph for the terminals of type i
        mainFrm->logMsg(tr("Processing %1 terminals of type %2").arg(terms[k].size()).arg(k+1));
        QCoreApplication::processEvents();

        CGraph_u gt;
        map<vDesc_u,vertex_data *> vmap;
        for(size_t j=0;j<terms[k].size();j++) {
            vDesc_u v=gt.insertVertex(terms[k][j]->id);
            vmap[v]=terms[k][j]->vd;
        }
        //Now insert edges if the distance between two vertices is below adjDist[k]

        for(map<vDesc_u,vertex_data *>::iterator it=vmap.begin();it!=vmap.end();++it) {
            int x=it->second->grid_coords.x;
            int y=it->second->grid_coords.y;
            vDesc_u ori=it->first;
            for(int i=x;i<=(int)(x+adjDist[k]);i++) {
                if(i>=(int)m) break;
                for(int j=(int)((double)y-adjDist[k]);j<=(int)(y+adjDist[k]);j++) {
                    if(j<0 || j>=(int)n || mat[i][j]==NULL) continue;


                    //Does the site belong to multiple types?
                    if(mat[i][j]->vd->siteType==numTypes+1) { //Yes
                        if(!binary_search(mat[i][j]->vd->cellTypes.begin(),mat[i][j]->vd->cellTypes.end(),k+1)) continue; //Not type k+1
                    }
                    else if(mat[i][j]->vd->siteType!=k+1) continue; //Not type k+1


                    if(calcDist(mat[x][y]->vd,mat[i][j]->vd)>adjDist[k]) continue;
                    vDesc_u dest=gt.findVertex(mat[i][j]->id).first;
                    if(ori==dest) continue;
                    gt.insertEdge(ori,dest,0);
                }
            }
        }
        //Calculate connected components
        vector< vector<vDesc_u> >vconn;
        gt.connComp(vconn);
        //Each connected component is a terminal of type k. Build a fake site for each

        for(size_t i=0;i<vconn.size();i++) {
            //insert the terminal number in the sites belonging to this terminal.Number terminals 1,2,...
            for(size_t j=0;j<vconn[i].size();j++) {
                vmap[vconn[i][j]]->termId[k]=contador;
            }
            fake_ids[contador]=maxId+contador;
            site_info *si=new site_info;
            si->id=maxId+contador;
            pa_ids[k].push_back(maxId+contador);
            vertex_data *ud=new vertex_data;
            si->vd=ud;
            ud->siteType=k+1;
            ud->cost=0;
            ud->friction=-1e+100;
            bool *B=new bool [numTypes];
            for(size_t j=0;j<numTypes;j++) B[j]=false;
            ud->barriers=B;
            ud->termId=(int *)contador; //No need to have an array
            fake_sites.push_back(si);
            vdMap[si->id]=ud;
            contador++;
        }
        mainFrm->logMsg(tr("\tFound %1 protected areas").arg(vconn.size()));
        QCoreApplication::processEvents();
    }

    mainFrm->logMsg(tr("Grid: %1x%2").arg(m).arg(n));
    QCoreApplication::processEvents();


    return true;
}


bool CData::insertEdge(int ori,int dest,CGraph_u *g) const
{
     pair<vDesc_u,bool> p=g->findVertex(ori);
     if(!p.second) {
         QMessageBox::critical(NULL,tr("Error"),tr("Vertex origin does not exist: %1").
                               arg(ori));
         return false;
     }
     vDesc_u vOri=p.first;

     p=g->findVertex(dest);
     if(!p.second) {
         QMessageBox::critical(NULL,tr("Error"),tr("Vertex target does not exist: %1").
                               arg(dest));
         return false;
     }
     vDesc_u vDest=p.first;

     //We rely in CGraph_u class to not insert duplicate edges.So do not teste here
     g->insertEdge(vOri,vDest,0);
    return true;
}


bool CData::buildGraph(CGraph_u **gk,double friction)
{


    //Build graphs for each type

    for(size_t k=0;k<numTypes;++k) {


        mainFrm->logMsg(tr("Building graph for type %1 with friction threshold=%2").arg(k+1).arg(friction));
        mainFrm->logMsg(tr("\tInserting vertices"));
        QCoreApplication::processEvents();

        //Non PA
        for(size_t i=0;i<sites.size();i++) {
            //Exclude PA of type k+1
            //Does the site belong to multiple types?
            if(sites[i]->vd->siteType==numTypes+1) { //Yes
                if(binary_search(sites[i]->vd->cellTypes.begin(),sites[i]->vd->cellTypes.end(),k+1)) continue; //Exclude PA of type k+1
            }
            else if(sites[i]->vd->siteType==k+1) continue; //Exclude PA of type k+1

            if(sites[i]->vd->friction <= friction && !sites[i]->vd->barriers[k]) {
                gk[k]->insertVertex(sites[i]->id);
            }
        }
        //PA
        for(size_t i=0;i<fake_sites.size();i++) {
            if(fake_sites[i]->vd->siteType!=k+1) continue;
            gk[k]->insertVertex(fake_sites[i]->id);
        }


        mainFrm->logMsg(tr("\tInserted %1 vertices").arg(gk[k]->size()));

        //Calculate adjacencies and insert the edges

        mainFrm->logMsg(tr("\tCalculating adjacencies and inserting edges"));
        QCoreApplication::processEvents();

        int ori=-1,dest=-1;
        for(int i=0;i<(int)m;i++) {
            for(int j=0;j<(int)n;j++) {
                if(mat[i][j]==NULL || mat[i][j]->vd->friction>friction || mat[i][j]->vd->barriers[k]) continue;

                if(mat[i][j]->vd->siteType==numTypes+1) {
                    if(binary_search(mat[i][j]->vd->cellTypes.begin(),mat[i][j]->vd->cellTypes.end(),k+1))
                        ori=fake_ids[mat[i][j]->vd->termId[k]];
                    else ori=mat[i][j]->id;
                }
                else if(mat[i][j]->vd->siteType==k+1) {
                    ori=fake_ids[mat[i][j]->vd->termId[k]];
                }
                else ori=mat[i][j]->id;

                for(int ii=(int)i;ii<=(int)(i+adjDist[k]);ii++) {
                    if(ii>=(int)m) break;
                    for(int jj=(int)((double)j-adjDist[k]);jj<=(int)(j+adjDist[k]);jj++) {
                        if(jj<0 || jj>=(int)n || mat[ii][jj]==NULL || mat[ii][jj]->vd->friction>friction || mat[ii][jj]->vd->barriers[k]) continue;


                        if(mat[ii][jj]->vd->siteType==numTypes+1) {
                            if(binary_search(mat[ii][jj]->vd->cellTypes.begin(),mat[ii][jj]->vd->cellTypes.end(),k+1))
                                dest=fake_ids[mat[ii][jj]->vd->termId[k]];
                            else dest=mat[ii][jj]->id;
                        }
                        else if(mat[ii][jj]->vd->siteType==k+1) {
                            dest=fake_ids[mat[ii][jj]->vd->termId[k]];
                        }
                        else dest=mat[ii][jj]->id;

                        if(ori==dest) continue;
                        double d=calcDist(mat[i][j]->vd,mat[ii][jj]->vd);
                        if(d>adjDist[k]) continue;
                        if(d<1e-16) continue; //sites with multiple types do not need to be adjacent
                        if(!insertEdge(ori,dest,gk[k])) {
                            return false;
                        }
                    }
                }

            }
        }

        mainFrm->logMsg(tr("\tInserted %1 edges").arg(gk[k]->numEdges()));
        QCoreApplication::processEvents();

        //Calculate connected components
        vector< vector<vDesc_u> >vconn;
        gk[k]->connComp(vconn);
        //count the number of components with 2 or more protected areas
        //sort descriptors in components
        for(size_t i=0;i<vconn.size();++i) sort(vconn[i].begin(),vconn[i].end());
        size_t count_comps=0,count_isolated=0;
        for(size_t i=0;i<vconn.size();++i) {
            size_t count=0;
            for(size_t j=0;j<fake_sites.size();j++) {
                if(fake_sites[j]->vd->siteType!=k+1) continue;
                vDesc_u v=gk[k]->findVertex(fake_sites[j]->id).first;
                if(binary_search(vconn[i].begin(),vconn[i].end(),v)) count ++;
                if(count==2) {count_comps++; break;}
            }
            if(count==1) count_isolated++;
        }
        mainFrm->logMsg(tr("\tFound %1 components with linkable PAs").arg(count_comps));
        if(count_isolated) mainFrm->logMsg(tr("\tFound %1 isolated PAs").arg(count_isolated));
        QCoreApplication::processEvents();

    }

    return true;
}

double CData::calcDist(vertex_data *v1,vertex_data *v2)
{
    double x1=v1->grid_coords.x,x2=v2->grid_coords.x;
    double y1=v1->grid_coords.y,y2=v2->grid_coords.y;
    return sqrt((x2-x1)*(x2-x1)+(y2-y1)*(y2-y1));
}
