/*
 * cgraph.h
 * Copyright (C) Raul Massano Brás 2012 <rbras@iseg.utl.pt>
 *
 * MulTyLink is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MulTyLink is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#ifndef __RB_GRAPH_H__
#define __RB_GRAPH_H__

const int RB_PATH_FOUND=133;


#include <boost/config.hpp>
#include <boost/graph/graph_traits.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/dijkstra_shortest_paths.hpp>
#include <boost/graph/prim_minimum_spanning_tree.hpp>
#include <boost/graph/connected_components.hpp>
#include <boost/graph/graph_utility.hpp>
#include <boost/graph/strong_components.hpp>


using namespace boost;
using namespace std;


//Undirected graphs


typedef adjacency_list_traits<setS, listS,
  undirectedS>::vertex_descriptor vDesc_u;

typedef adjacency_list < setS, listS, undirectedS,
  property<vertex_index_t, int,
  property<vertex_distance_t, double,
  property<vertex_predecessor_t, vDesc_u> > >,
  property < edge_weight_t, double > > Graph_u;

typedef graph_traits<Graph_u>::vertex_iterator vIter_u;
typedef graph_traits<Graph_u>::edge_descriptor eDesc_u;
typedef graph_traits<Graph_u>::edge_iterator eIter_u;
typedef graph_traits<Graph_u>::adjacency_iterator aIter_u;

struct caminho_u_t
{
  list<vDesc_u> l;
  double cost;
};

class CGraph_u {
 private:
  Graph_u g;
  map<vDesc_u,size_t> idMap;
  map<size_t,vDesc_u> id2idx;
 public:
  CGraph_u() {}
  vDesc_u insertVertex(size_t id) { vDesc_u v=add_vertex(g); idMap[v]=id; id2idx[id]=v; return v;}
  void removeVertex(vDesc_u idx) {
    size_t id=idMap[idx]; idMap.erase(idx); id2idx.erase(id);
    clear_vertex(idx,g); remove_vertex(idx,g);
  }

  bool insertEdge(vDesc_u ori,vDesc_u dest, double c) {eDesc_u e;bool flag; tie(e,flag)=add_edge(ori,dest,c,g); return flag;}
  vDesc_u getEdgeSource(eDesc_u e) {return source(e,g);}
  vDesc_u getEdgeTarget(eDesc_u e) {return target(e,g);}
  size_t getId(vDesc_u v) {return idMap[v];}
  void setEdgeCost(eDesc_u e, double x) {
      property_map<Graph_u, edge_weight_t>::type cost=get(edge_weight, g);
      put(cost,e,x);
  }
  double getEdgeCost(eDesc_u e) {
      property_map<Graph_u, edge_weight_t>::type cost = get(edge_weight, g);
      return get(cost,e);
  }

  size_t getDegree(vDesc_u u) {return degree(u,g);}
  size_t size() {return num_vertices(g);}
  size_t numEdges() {return num_edges(g);}
  
  pair<aIter_u,aIter_u> getAdjVertices(vDesc_u v) {return adjacent_vertices(v,g);} 
  pair<vIter_u,vIter_u> getVertices() {return vertices(g);}
  pair<eIter_u,eIter_u> getEdges() {return edges(g);}
  pair<eDesc_u,bool> getEdge(vDesc_u u,vDesc_u v) {return edge(u,v,g);}


  pair<vDesc_u,bool> findVertex(size_t id) {
    pair<vDesc_u,bool> p;
    p.second=false;
    map<size_t,vDesc_u>::iterator it;
    if((it=id2idx.find(id))==id2idx.end()) return p;
    p.second=true;
    p.first=it->second;
    return p;
  }

  void connComp(vector< vector<vDesc_u> >& vconn)
  {
    //Calculate connected components
    //Return a vector with the n components. Each component is saved in a vactor

    vconn.clear();

    //fill the indices

    property_map<Graph_u, vertex_index_t>::type index = get(vertex_index, g);

    graph_traits<Graph_u>::vertices_size_type cnt = 0;

    vIter_u vi, vend;
    for(tie(vi,vend) = vertices(g); vi != vend; ++vi) put(index, *vi, cnt++);

    vector<int> cc(num_vertices(g));

    size_t num_comps=connected_components(g,make_iterator_property_map(cc.begin(),get(vertex_index,g)));

    for(size_t i=0;i<num_comps;i++) {
      vector<vDesc_u> v;
      for(tie(vi,vend) = vertices(g); vi != vend; ++vi)
	if(cc[index[*vi]]==(int)i) v.push_back(*vi); 
      vconn.push_back(v);
    }

  }

  //Minimum Spanning Tree
  void prim(vDesc_u ori,vector< pair<vDesc_u,vDesc_u> >& mst)
  {
    mst.clear();

    property_map<Graph_u, vertex_index_t>::type indexmap = get(vertex_index, g);
    property_map<Graph_u, edge_weight_t>::type weightmap = get(edge_weight, g);
    
    vIter_u vi,vi_end;
    int i = 0;
    for (tie(vi,vi_end)=vertices(g);vi!=vi_end;++vi,++i) {
      indexmap[*vi]=i;
    }
    
    property_map<Graph_u, vertex_predecessor_t>::type p=get(vertex_predecessor, g);
    property_map<Graph_u, vertex_distance_t>::type d = get(vertex_distance, g);
    
    
    prim_minimum_spanning_tree(g, ori, p, d,weightmap, indexmap,default_dijkstra_visitor());
    
    
    for (tie(vi,vi_end)=vertices(g);vi!=vi_end;++vi) {
      if(p[*vi]==*vi) continue; //Not in the same cpmponent
      mst.push_back(make_pair(*vi,p[*vi]));
    }

  }

  void printGraph()
  {
      property_map<Graph_u, vertex_index_t>::type index = get(vertex_index, g);

      vIter_u vi, vend;
      for(tie(vi,vend) = vertices(g); vi != vend; ++vi) put(index, *vi, idMap[*vi]);

      print_graph(g);

  }

};

//Directed graphs


typedef adjacency_list_traits<setS, listS,
  undirectedS>::vertex_descriptor vDesc_d;

typedef adjacency_list < setS, listS, directedS,
  property<vertex_index_t, int,
  property<vertex_distance_t, double,
  property<vertex_predecessor_t, vDesc_d> > >,
  property < edge_weight_t, double > > Graph_d;

typedef graph_traits<Graph_d>::vertex_iterator vIter_d;
typedef graph_traits<Graph_d>::edge_descriptor eDesc_d;
typedef graph_traits<Graph_d>::edge_iterator eIter_d;
typedef graph_traits<Graph_d>::adjacency_iterator aIter_d;



struct caminho_d_t
{
  list<vDesc_d> l;
  double cost;
};


//definição da classe

class CGraph_d {
 private:
  Graph_d g;
  map<vDesc_d,size_t> idMap;
  map<size_t,vDesc_d> id2idx;
 public:
  CGraph_d() {}


  vDesc_d insertVertex(size_t id) { vDesc_d v=add_vertex(g); idMap[v]=id; id2idx[id]=v; return v;}
  void removeVertex(vDesc_d idx) {
    size_t id=idMap[idx]; idMap.erase(idx); id2idx.erase(id);
    clear_vertex(idx,g); remove_vertex(idx,g);
  }

  void insertEdge(vDesc_d ori,vDesc_d dest, double c) {add_edge(ori,dest,c,g);}
  vDesc_d getEdgeSource(eDesc_d e) {return source(e,g);}
  vDesc_d getEdgeTarget(eDesc_d e) {return target(e,g);}
  size_t getId(vDesc_d v) {return idMap[v];}
  void setEdgeCost(eDesc_d e, double x) {
      property_map<Graph_d, edge_weight_t>::type cost=get(edge_weight, g);
      put(cost,e,x);
  }
  double getEdgeCost(eDesc_d e) {
      property_map<Graph_d, edge_weight_t>::type cost = get(edge_weight, g);
      return get(cost,e);
  }


  size_t size() {return num_vertices(g);}
  size_t numEdges() {return num_edges(g);}

  pair<aIter_d,aIter_d> getAdjVertices(vDesc_d v) {return adjacent_vertices(v,g);} 
  pair<vIter_d,vIter_d> getVertices() {return vertices(g);}
  pair<eIter_d,eIter_d> getEdges() {return edges(g);}
  pair<eDesc_d,bool> getEdge(vDesc_d u,vDesc_d v) {return edge(u,v,g);}

  pair<vDesc_d,bool> findVertex(size_t id) {
    pair<vDesc_d,bool> p;
    p.second=false;
    map<size_t,vDesc_d>::iterator it;
    if((it=id2idx.find(id))==id2idx.end()) return p;
    p.second=true;
    p.first=it->second;
    return p;
  }

  //Shortest Paths
  void dijkstra(vDesc_d ori,const vector<vDesc_d>& dest, vector<caminho_d_t>& c)
  {

    //On entry vector dest contains descriptors of the vertices for wich we return
    //the shortest path and the cost in vector c

    c.clear();

    //property_map<Graph_d, edge_weight_t>::type weightmap = get(edge_weight, g);
    property_map<Graph_d, vertex_index_t>::type indexmap = get(vertex_index, g);
    vIter_d vi,vi_end;
    int i = 0;
    for (tie(vi,vi_end)=vertices(g);vi!=vi_end;++vi,++i) {
      indexmap[*vi]=i;
    }

    property_map<Graph_d, vertex_distance_t>::type d=get(vertex_distance, g);
    property_map<Graph_d, vertex_predecessor_t>::type p=get(vertex_predecessor, g);

    dijkstra_shortest_paths(g, ori, predecessor_map(p).distance_map(d));


    for(size_t i=0;i<dest.size();i++) {

      list<vDesc_d> lista;
      vDesc_d u=dest[i];
      lista.push_front(u);
      do {
	if(u==p[u]) break;
	u=p[u];
	lista.push_front(u);
      }while(u!=ori);
      if(lista.size()==1) lista.clear(); //No path
      caminho_d_t r={lista,d[dest[i]]};
      c.push_back(r);
    }
  
  }


  class VisitorDestUnique : public default_dijkstra_visitor {
    public:
       VisitorDestUnique(vDesc_d v_) : v(v_) {}
       template <typename Vertex, typename Graph>
         void examine_vertex(Vertex u, Graph& ) {
         if(u==v) throw(RB_PATH_FOUND);
    }
    private:
       vDesc_d v;
    
  };


  void dijkstra_dest_unique(vDesc_d ori,vDesc_d dest, caminho_d_t& c)
  {
    c.l.clear();

    //property_map<Graph_d, edge_weight_t>::type weightmap = get(edge_weight, g);
    property_map<Graph_d, vertex_index_t>::type indexmap = get(vertex_index, g);
    vIter_d vi,vi_end;
    int i = 0;
    for (tie(vi,vi_end)=vertices(g);vi!=vi_end;++vi,++i) {
      indexmap[*vi]=i;
    }

    property_map<Graph_d, vertex_distance_t>::type d=get(vertex_distance, g);
    property_map<Graph_d, vertex_predecessor_t>::type p=get(vertex_predecessor, g);

    VisitorDestUnique vis(dest);


    try {
      dijkstra_shortest_paths(g, ori, predecessor_map(p).distance_map(d).visitor(vis));
    }
    catch(int e) {
      if(e==RB_PATH_FOUND) {
	vDesc_d u=dest;
	c.l.push_front(u);
	do {
	  if(u==p[u]) break;
	  u=p[u];
	  c.l.push_front(u);
	}while(u!=ori);
	if(c.l.size()==1) c.l.clear(); //No path
        c.cost=d[dest];
      }
    }
  }




  class VisitorDestFirst : public default_dijkstra_visitor {
    public:
       VisitorDestFirst(const vector<vDesc_d>& v_,vDesc_d& d_) : v(v_),d(d_) {}
       template <typename Vertex, typename Graph> void examine_vertex(Vertex u, Graph& ) {
	 if(binary_search(v.begin(),v.end(),u)) {
	   d=u;
           throw(RB_PATH_FOUND);
	 }
       }
    private:
       const vector<vDesc_d>& v;
       vDesc_d& d;
  };



  void dijkstra_first(vDesc_d ori,const vector<vDesc_d>& dest, caminho_d_t& c,vDesc_d& first)
  {
    //Returns the path to the first vertex found in dest
    //dest must be sorted!
    c.l.clear();

    //property_map<Graph_d, edge_weight_t>::type weightmap = get(edge_weight, g);
    property_map<Graph_d, vertex_index_t>::type indexmap = get(vertex_index, g);
    vIter_d vi,vi_end;
    int i = 0;
    for (tie(vi,vi_end)=vertices(g);vi!=vi_end;++vi,++i) {
      indexmap[*vi]=i;
    }

    property_map<Graph_d, vertex_distance_t>::type d=get(vertex_distance, g);
    property_map<Graph_d, vertex_predecessor_t>::type p=get(vertex_predecessor, g);


    
    first=NULL;
    VisitorDestFirst vis(dest,first);

    try {
      dijkstra_shortest_paths(g, ori, predecessor_map(p).distance_map(d).visitor(vis));
    }
    catch(int e) {
      if(e==RB_PATH_FOUND) {
        vDesc_d u=first;
	c.l.push_front(u);
	do {
	  if(u==p[u]) break;
	  u=p[u];
	  c.l.push_front(u);
	}while(u!=ori);
	if(c.l.size()==1) c.l.clear(); //No path
        c.cost=d[first];
      }
    }
  }

  //Strong components of a directed graph

  void strongConnComp(vector< vector<vDesc_d> >& vconn)
  {
    //Calculate strong connected components
    //Return a vector with the n components. Each component is saved in a vector

    vconn.clear();

    //fill the indices

    property_map<Graph_d, vertex_index_t>::type index = get(vertex_index, g);

    graph_traits<Graph_d>::vertices_size_type cnt = 0;

    vIter_d vi, vend;
    for(tie(vi,vend) = vertices(g); vi != vend; ++vi) put(index, *vi, cnt++);

    vector<int> cc(num_vertices(g));

    size_t num_comps=strong_components(g,make_iterator_property_map(cc.begin(),get(vertex_index,g)));

    for(size_t i=0;i<num_comps;i++) {
      vector<vDesc_d> v;
      for(tie(vi,vend) = vertices(g); vi != vend; ++vi)
	if(cc[index[*vi]]==(int)i) v.push_back(*vi); 
      vconn.push_back(v);
    }

  }
  

};



#endif
