/*
 * cproblem.h
 * Copyright (C) Raul Massano Brás 2012 <rbras@iseg.utl.pt>
 *
 * MulTyLink is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MulTyLink is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef CPROBLEM_H
#define CPROBLEM_H

#include "cgraph.h"
#include <vector>
#include <set>
#include <string>
#include <QString>
#include <QPoint>
#include <cmath>
#include <map>
#include "mainwindow.h"

class MainWindow;

struct graphSol { //return a solution
    std::vector<int> vertices;
    double cost;
};


struct point {
    point(size_t i=0, size_t j=0):x(i),y(j){}
    size_t x;
    size_t y;
};

struct vertex_data {
    point grid_coords;
    double xcoord,ycoord;
    unsigned char siteType; //0 is a general site, 1,...K is a terminal of type k
    bool *barriers; //barrier[k]=true means that the site is a barrier for type k
    double cost;    //site cost
    double friction; //if this value is above the limit specified by the user in the parametres the site
                     //is not used
    int *termId;     //Id of the terminals in each type. It is NULL if the site is not a terminal
    vector<int> cellTypes; //If the cell belongs to severall types save them here
};

typedef map<int,vertex_data *> vdMap_t;

struct params {
    double limit; //sites with friction over this value are not used
    size_t nt; //number of threads to solve the permutations
    size_t nperm; //number of permutations to test
    double epsilon;
    int numRep; //number of repetitions for the greedy algorithm
    int time_limit; //Time limit in seconds for the greedy algorithm
};

struct site_info
{
    int id;
    vertex_data *vd;
};

struct problemSol
{
    graphSol *sol;
    double *cost;
    double *min_friction;
    double *max_friction;
    double *min_cost;
    double *max_cost;
    double totCost;
    size_t *nSites;
    size_t totSites;
    size_t *numComps;
};

class CData;
class CScene;
int binSearch(const std::vector<int> x, int val);

class CProblem : public QObject
{
private:
    size_t numTypes;  //Number of different reserve types
    std::string lastError;
    params pars;
    QPoint grid;
    CData *data;   //store a pointer to the data

    size_t zoom;

    CGraph_u **gk;

    MainWindow *mainFrm;

public:
    CProblem(MainWindow *mainFrm);
    ~CProblem();
    bool readData(const QString& file);

    const QPoint& getGrid() const {return grid;}

    size_t getNumTypes() const {return numTypes;}
    const std::string& getLastError() const {return lastError;}
    //const std::string& getName(size_t i) const {return names[i];}
    const params *getParams() const {return &pars;}
    void setParams(double friction_limit, size_t nt,size_t np,int nr, int tl);

    const std::set<vertex_data *>& getFrontier(size_t i, size_t j) const;
    size_t getNumPa(size_t i) const;
    const site_info *getData(int i, int j) const;
    CData *getDataPtr() const {return data;}
    size_t getZoom() {return zoom;}
    void setZoom(int val) {zoom=val;}
    const MainWindow *getMainFrm() {return mainFrm;}
    CGraph_u **getGraphTypePtr() {return gk;}

private:
    void errMsg(const std::string& str) {lastError=str;}
};

#endif // CPROBLEM_H
