/*
 * cscene.cpp
 * Copyright (C) Raul Massano Brás 2012 <rbras@iseg.utl.pt>
 *
 * MulTyLink is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MulTyLink is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "cscene.h"
#include "version.h"
#include "stringutils.h"
#include "csetparams.h"
#include "ui_cscene.h"
#include <QMessageBox>
#include <QFileDialog>
#include <QPrinter>
#include <QPrintDialog>
#include <QPainter>
#include <QTextStream>
#include <fstream>
#include <set>
#include <iostream>
#include "cproblem.h"
#include "cdata.h"
#include "cthreads.h"
#include <algorithm>
#include "cthreadsgreedy.h"
#include "csetcolours.h"

using namespace std;

static const size_t max_default_colours=10;
static const QColor pa_type_default_colour[max_default_colours]={Qt::green,Qt::red,Qt::blue,Qt::magenta,Qt::cyan,Qt::darkCyan,Qt::darkGreen,Qt::darkBlue,
                                                        Qt::darkMagenta,Qt::darkYellow};
static const QColor multiTypeColour=Qt::darkRed;
QColor CScene::baseColors[COLOR_END]={Qt::darkRed,QColor(9,138,186,255),Qt::white,Qt::yellow,Qt::black,Qt::gray};

CScene::CScene(QWidget *parent,CProblem *pr) :
    QMainWindow(parent),
    m_ui(new Ui::CScene)
{
    m_ui->setupUi(this);
    mainWin=parent;

    problem=pr;
    label=m_ui->label;
    bmap=NULL;
    newSol=false;
    solProb=NULL;

    pb=m_ui->progressBar;
    clabel=m_ui->label_cost;
    popup=NULL;
    thread=NULL;
    time.start();
    checkedTypes=new bool [problem->getNumTypes()];
    checkedTypesBarr=new bool [problem->getNumTypes()];
    m_ui->actionAll->setVisible(false);
    m_ui->actionAllBarr->setVisible(false);

    act=new QAction* [problem->getNumTypes()];
    actBarr=new QAction* [problem->getNumTypes()];
    for(size_t i=0;i<problem->getNumTypes();i++) {
        act[i]=m_ui->menuPA_Types->addAction(QString(problem->getDataPtr()->getPAName(i).c_str()));
        act[i]->setCheckable(true);
        act[i]->setChecked(true);
        checkedTypes[i]=true;

         //newAct->setShortcuts(QKeySequence::New);
         //newAct->setStatusTip(tr("Create a new file"));
        connect(act[i], SIGNAL(triggered()),this, SLOT(on_actionAll_triggered()));
        actBarr[i]=m_ui->menuBarriers->addAction(QString(problem->getDataPtr()->getPAName(i).c_str()));
        actBarr[i]->setCheckable(true);
        actBarr[i]->setChecked(false);
        checkedTypesBarr[i]=false;

         //newAct->setShortcuts(QKeySequence::New);
         //newAct->setStatusTip(tr("Create a new file"));
        connect(actBarr[i], SIGNAL(triggered()),this, SLOT(on_actionAllBarr_triggered()));
    }
    typeChanged=false; colorChanged=false;
    firstView=true;
    setZoomValue(problem->getZoom());

    typeColor=new QColor [problem->getNumTypes()];
    for(size_t i=0;i<problem->getNumTypes();i++) typeColor[i]=pa_type_default_colour[i%max_default_colours];
}


CScene::~CScene()
{
    cleanSol();
    if(bmap) delete bmap;
    delete [] act;
    delete [] actBarr;
    delete [] checkedTypes;
    delete [] checkedTypesBarr;
    delete [] typeColor;
    delete m_ui;
}

void CScene::mousePressEvent(QMouseEvent *ev)
{
    QPoint p;
    p=label->mapFromGlobal(ev->globalPos());
    if(p.x()<0 || p.y()<0 || p.x()>label->width() || p.y()>label->height()) return;

    size_t zoom=problem->getZoom();

    size_t cx=p.x()/zoom,cy=p.y()/zoom;

    const site_info *si=problem->getData(cx,cy);
    if(si) {
        QString str;
        if(si->vd->siteType<=problem->getNumTypes()) str.setNum(si->vd->siteType);
        else {
            str.setNum(si->vd->cellTypes[0]);
            for(size_t i=1;i<si->vd->cellTypes.size();i++) {
                QString str1;
                str1=tr(",%1").arg(si->vd->cellTypes[i]);
                str+=str1;
            }
        }
        popup=new PopupMessage(this,QString("id=%1\nsite type=%2\ncost=%3\nx=%4\ny=%5").arg(si->id).
                               arg(str).arg(si->vd->cost).arg(si->vd->xcoord).arg(si->vd->ycoord));
        popup->move(x()+p.x(),y()+p.y());
        popup->show();
        //QMessageBox msgBox(QMessageBox::Information,tr("Info"),tr("id=%1\npa_id=%2\npa_type=%3").arg(si->id).arg(si->vd->PA_id).
        //                   arg(si->vd->PA_type!=-1?si->vd->PA_type+1:-1));
        //msgBox.exec();
    }
    ev->accept();
}

void CScene::mouseReleaseEvent(QMouseEvent *ev)
{
    if(popup) {
        delete popup;
        popup=NULL;
        ev->accept();
    }
}

void CScene::changeEvent(QEvent *e)
{
    QMainWindow::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
        m_ui->retranslateUi(this);
        break;
    default:
        break;
    }
}


void CScene::on_actionClose_triggered()
{
    close();
}

void CScene::closeEvent(QCloseEvent *event)
 {
     event->accept();
     emit(winCloseSignal(this));
 }

void CScene::on_actionSave_as_triggered()
{
    if(!solProb) return;

    QString file = QFileDialog::getSaveFileName(this,tr("File to save solution"),
                                                ".", tr("Text files (*.txt)"));

    if(file.isEmpty()) return;
    if(file.mid(file.lastIndexOf(".")+1).toLower()!="txt") file+=".txt";

    QFile f(file);
    if(!f.open(QIODevice::WriteOnly | QIODevice::Text)) {
        QMessageBox::critical(mainWin,QObject::tr("Error"),QObject::tr("Could not open %1 for writing").arg(file));
        return;
    }

    QTextStream out(&f);
    out.setRealNumberPrecision(9);
    //Write the type of problem
    const params *pars=problem->getParams();
    CData *data=problem->getDataPtr();

    out << tr("\nSolution with friction threshold=%1\n\n").
                    arg(pars->limit);


    for(size_t i=0;i<problem->getNumTypes();i++) {
        out << trUtf8("\nSolution for %1\n").arg(QString(data->getPAName(i).c_str()));
        for(size_t j=0;j<solProb->sol[i].vertices.size();j++) {
            vertex_data *vd=data->getVertexDataById(solProb->sol[i].vertices[j]);
            out << solProb->sol[i].vertices[j] << ',' << vd->xcoord << ',' << vd->ycoord << endl;
        }
    }
    out << "\n";

    for(size_t i=0;i<problem->getNumTypes();i++) {
        out << trUtf8("\nType %1:\n").arg(QString(data->getPAName(i).c_str()));
        out << trUtf8("\tCost:           %1\n").arg(solProb->cost[i]);
        out << trUtf8("\tNum sites:      %1\n").arg(solProb->nSites[i]);
        out << trUtf8("\tMin. friction: %1\n").arg(solProb->min_friction[i]);
        out << trUtf8("\tMax. friction: %1\n").arg(solProb->max_friction[i]);
        out << trUtf8("\tMin. cost:    %1\n").arg(solProb->min_cost[i]);
        out << trUtf8("\tMax. cost:    %1\n").arg(solProb->max_cost[i]);
        out << trUtf8("\tNum. comps.:    %1\n").arg(solProb->numComps[i]);
    }
    out << trUtf8("\nTotals:\n");

    out << trUtf8("\tCost:           %1\n").arg(solProb->totCost);
    out << trUtf8("\tNum sites       %1\n").arg(solProb->totSites);
    out << trUtf8("\tMin. friction: %1\n").arg(*min_element(solProb->min_friction,solProb->min_friction+problem->getNumTypes()));
    out << trUtf8("\tMax. friction: %1\n").arg(*max_element(solProb->max_friction,solProb->max_friction+problem->getNumTypes()));
    out << trUtf8("\tMin. cost:      %1\n").arg(*min_element(solProb->min_cost,solProb->min_cost+problem->getNumTypes()));
    out << trUtf8("\tMax. cost:      %1\n").arg(*max_element(solProb->max_cost,solProb->max_cost+problem->getNumTypes()));


}


void CScene::on_actionSet_Params_triggered()
{
    const params *pars=problem->getParams();
    QStringList typeNames;
    for(size_t i=0;i<problem->getNumTypes();i++) typeNames << QString(problem->getDataPtr()->getPAName(i).c_str());
    CSetParams dlg(this,pars,problem->getNumTypes(),problem->getDataPtr()->getMaxFriction());
    if(dlg.exec()==QDialog::Accepted) {
        size_t nt,np;
        double  friction_limit;
        int nr,tl;
        dlg.getVals(friction_limit,nt,np,nr,tl);
        double  old_limit=pars->limit;
        enableMenus(false);
        problem->setParams(friction_limit,nt,np,nr,tl);
        if(friction_limit!=old_limit) {
            cleanSol();
            newSol=true;
            update();
        }
        enableMenus(true);
    }
}

void CScene::on_actionSolve_2_triggered()
{
    if(solProb) { //Clear current solution
        cleanSol();
        setProgress(0);
        clabel->setText(QString(tr("Cost: 0.0")));
        update();
    }
    newSol=false;
    if(!solProb) {
        size_t K=problem->getNumTypes();
        solProb=new problemSol;
        solProb->sol=new graphSol[K];
        solProb->cost=new double [K];
        solProb->min_friction=new double [K];
        solProb->max_friction=new double [K];
        solProb->min_cost=new double [K];
        solProb->max_cost=new double [K];
        solProb->nSites=new size_t[K];
        solProb->numComps=new size_t[K];
    }

    for(size_t i=0;i<problem->getNumTypes();i++) solProb->sol[i].vertices.clear();


    thread=new CSolveThread(this,problem,solProb);
    connect(thread,SIGNAL(threadMsgSignal(int, uint, logMsg_s )),SLOT(threadMessages(int, uint, logMsg_s)));
    connect(thread,SIGNAL(threadProgressSignal(int)),SLOT(threadProgressMessages(int)));

    time.restart();
    thread->start();
    enableMenus(false);

    //problem->solve(solProb,this);
 }

void CScene::paintEvent(QPaintEvent *ev)
{

    static uint zoomAnt=0;

    size_t zoom=problem->getZoom();

    if(!problem) return; //problem by have been reset

    QPoint grid=problem->getGrid();

    if(newSol || zoom!=zoomAnt ||  typeChanged || firstView || colorChanged) {
        firstView=false;
        //Create a new bitmap
        if(bmap) delete bmap;
        bmap=new QPixmap(grid.x(),grid.y());
        bmap->fill(Qt::lightGray);
        //Draw in the pixmap
        QPainter dc;
        dc.begin(bmap);
        DrawImage(&dc);
        dc.end();

        //Apply zoom
        if(zoom!=1) {
            QPixmap *pix=new QPixmap((int)(grid.x()*zoom),(int)(grid.y()*zoom));
            *pix=bmap->scaled((int)(grid.x()*zoom),(int)(grid.y()*zoom));
            delete bmap;
            bmap=pix;
        }

        zoomAnt=zoom;
        newSol=false;
        typeChanged=false;
        colorChanged=false;

        label->setPixmap(*bmap);

    }
    ev->accept();
}


void CScene::DrawImage(QPainter *dc)
{

    CData *data=problem->getDataPtr();
    const params *pars=problem->getParams();
    QPoint pt=data->getGrid();
    size_t K=problem->getNumTypes();

    dc->setPen(baseColors[COLOR_Background]);
    dc->fillRect(QRect(0,0,pt.x(),pt.y()),baseColors[COLOR_Background]);
    /*
    for(int i=0;i<pt.x();i++)
        for(int j=0;j<pt.y();j++)
            if(data->getData(i,j)==NULL) dc->drawPoint(i,j);
    */
    for(size_t i=0;i<data->getNumSites();i++) { //for each site
        const vertex_data *vd=data->getSiteData(i)->vd;
        if(vd->friction>pars->limit) dc->setPen(baseColors[COLOR_Friction]);
        else if(vd->siteType==0) dc->setPen(baseColors[COLOR_NoType]);
        else if(vd->siteType==K+1) {
            //This is a site with multiple types. If it has only one of the selected
            //types we want to see set its colour, otherwise set the multicolour
            int count=0,type=-1;
            for(size_t j=0;j<K;j++) {
                if(checkedTypes[j] && binary_search(vd->cellTypes.begin(),vd->cellTypes.end(),j+1)) {count++; type=j;}
            }
            if(count==1) dc->setPen(typeColor[type]);
            else dc->setPen(baseColors[COLOR_MultiType]);
        }
        else dc->setPen(typeColor[vd->siteType-1]);

        //If we want to see barriers change colour here
        for(size_t j=0;j<problem->getNumTypes();j++) {
            if( checkedTypesBarr[j] && vd->barriers[j]) {
                dc->setPen(baseColors[COLOR_Barriers]);
                break;
            }
        }
        dc->drawPoint(vd->grid_coords.x,vd->grid_coords.y);
    }

    if(solProb) {
        dc->setPen(baseColors[COLOR_Solution]);
        int siteType;
        for(size_t i=0;i<K;i++) {
            if(!checkedTypes[i]) continue;
            for(size_t j=0;j<solProb->sol[i].vertices.size();j++) {
                vertex_data *vd=data->getVertexDataById(solProb->sol[i].vertices[j]);

                siteType=vd->siteType;
                //if the site is a terminal and its type is selected do not draw inside it

                if(siteType!=0) {
                    if(siteType<=(int)K && checkedTypes[siteType-1]) continue; //terminal with only one type
                    if(siteType==(int)K+1) {
                        //Here we have a multi type terminal. Do not draw inside if it has any of
                        //the checked types
                        bool noDraw=false;
                        for(size_t k=0;k<K;k++) {
                            if(checkedTypes[k] && binary_search(vd->cellTypes.begin(),vd->cellTypes.end(),k+1)) {
                                noDraw=true;
                                break;
                            }
                        }
                        if(noDraw) continue;
                    }
                }
                dc->drawPoint(vd->grid_coords.x,vd->grid_coords.y);
            }
        }
    }
}


void CScene::on_actionSave_Image_triggered()
{
    if(!bmap) return;

    QString filter;
    QString file = QFileDialog::getSaveFileName(this,tr("File to save image"),
                                                ".", tr("TIFF (*.tif);;JPEG (*.jpg);;EPS (*.eps);;Windows bitmap (*.bmp)"),&filter);

    if(file.isEmpty()) return;


    if(filter=="TIFF (*.tif)") {
        if(file.mid(file.lastIndexOf(".")+1).toLower()!="tif") file+=".tif";
    }
    else if(filter=="JPEG (*.jpg)") {
        if(file.mid(file.lastIndexOf(".")+1).toLower()!="jpg") file+=".jpg";
    }
    else if(filter=="Windows bitmap (*.bmp)") {
        if(file.mid(file.lastIndexOf(".")+1).toLower()!="bmp") file+=".bmp";
    }
    else if(filter=="EPS (*.eps)") {
        if(file.mid(file.lastIndexOf(".")+1).toLower()!="eps") file+=".eps";
        // Setup printer

        QPrinter printer(QPrinter::HighResolution);
        printer.setOutputFormat(QPrinter::PostScriptFormat);
        printer.setOutputFileName(file);
        printer.setFullPage(true);
        //printer.setPaperSize(QPrinter::A4);
        printer.setPaperSize(QSizeF(bmap->widthMM(), bmap->heightMM()), QPrinter::Millimeter);
        //printer.setOrientation(QPrinter::Landscape);
        printer.setPageMargins(0,0,0,0,QPrinter::Millimeter);
        // Get the painting context from the printer
        QPainter painter(&printer); // Destructing painter writes the  file
        // Draw the image
        double  scale = printer.width()/bmap->width();
        double  scale1 = printer.height()/bmap->height();
        painter.setTransform(QTransform(scale,0,0, 0,scale1,0, 0,0,1.0));
        QRectF src(0,0,bmap->width(),bmap->height());
        QRectF dst(0,0,bmap->width(),bmap->height());
        painter.drawPixmap(dst,*bmap,src);
        return;
    }
    else {
        QMessageBox::critical(0,tr("Filter error"),tr("Invalid filter:\n")+filter+"\n");
        return;
    }
    if(!bmap->save(file)) {
        QMessageBox::critical(0,tr("Format error"),tr("Could not create file.\nPlease verify if you have the dynamic libraries for the format specified."));
    }

}

void CScene::cleanSol()
{
    if(solProb) {
        delete [] solProb->sol;
        delete [] solProb->cost;
        delete [] solProb->nSites;
        delete [] solProb->min_friction;
        delete [] solProb->min_cost;
        delete [] solProb->max_friction;
        delete [] solProb->max_cost;
        delete [] solProb->numComps;
        delete solProb;
    }
    solProb=NULL;

}
void CScene::threadMessages(int m, uint p1, logMsg_s p2)
{
    if(m & MSG_FINISHED) {

        thread->wait();
        disconnect(thread,0,0,0);
        delete thread;
        thread=NULL;
        for(size_t i=0;i<problem->getNumTypes();i++) {
            if(act[i]->isChecked()) checkedTypes[i]=true;
            else checkedTypes[i]=false;
        }

        typeChanged=true;
        update();

        ((MainWindow *)mainWin)->logMsg(tr("Elapsed time: %1 seconds").arg(time.elapsed()/1000.0));
        //double  cost=0;
        //for(size_t i=0;i<problem->getNumTypes();i++) cost+=solProb->sol[i].cost;

        clabel->setText(QString(tr("Cost: %1").arg(solProb->totCost)));

        newSol=true;
        update();
        enableMenus(true);
    }
    if(m & MSG_PROGRESS) setProgress(p1);
    if(m & MSG_LOG) ((MainWindow *)mainWin)->logMsg(p2.msg);
    if(m & MSG_LABEL) {
        m_ui->label_task->setText(p2.msg);
        setProgress(p1);
    }
}

void CScene::threadProgressMessages(int p1)
{
    setProgress(p1);
}

void CScene::enableMenus(bool f)
{
    m_ui->actionClose->setEnabled(f);
    m_ui->actionSave_as->setEnabled(f);
    m_ui->actionSave_Image->setEnabled(f);
    m_ui->actionSet_Params->setEnabled(f);
    m_ui->actionSolve_2->setEnabled(f);
    m_ui->actionSolve_using_Greedy->setEnabled(f);
}

void CScene::on_actionAll_triggered()
{
    for(size_t i=0;i<problem->getNumTypes();i++) {
        if(act[i]->isChecked()) checkedTypes[i]=true;
        else checkedTypes[i]=false;
    }

    typeChanged=true;
    update();
}

void CScene::on_actionAllBarr_triggered()
{
    for(size_t i=0;i<problem->getNumTypes();i++) {
        if(actBarr[i]->isChecked()) checkedTypesBarr[i]=true;
        else checkedTypesBarr[i]=false;
    }

    typeChanged=true;
    update();
}


void CScene::on_horizontalSlider_sliderMoved(int position)
{
    int zoom=(int)(position/100.0+0.5);
    problem->setZoom(zoom);
    update();
}

void CScene::setZoomValue(int val)
{
    m_ui->horizontalSlider->setValue(100*val);
    m_ui->label_zoom->setText(QString("%1").arg(100*val));
}

void CScene::on_actionSolve_using_Greedy_triggered()
{
    if(solProb) { //Clear current solution
        cleanSol();
        setProgress(0);
        clabel->setText(QString(tr("Cost: 0.0")));
        update();
    }
    newSol=false;
    if(!solProb) {
        size_t K=problem->getNumTypes();
        solProb=new problemSol;
        solProb->sol=new graphSol[K];
        solProb->cost=new double [K];
        solProb->min_friction=new double [K];
        solProb->max_friction=new double [K];
        solProb->min_cost=new double [K];
        solProb->max_cost=new double [K];
        solProb->nSites=new size_t[K];
        solProb->numComps=new size_t[K];
    }

    for(size_t i=0;i<problem->getNumTypes();i++) solProb->sol[i].vertices.clear();


    thread=new CThreadsGreedy(this,problem,solProb);
    connect(thread,SIGNAL(threadMsgSignal(int, uint, logMsg_s )),SLOT(threadMessages(int, uint, logMsg_s)));
    connect(thread,SIGNAL(threadProgressSignal(int)),SLOT(threadProgressMessages(int)));

    time.restart();
    thread->start();
    enableMenus(false);


}

void CScene::on_actionColors_triggered()
{
    size_t K=problem->getNumTypes();
    csetcolours dlg(this,K,baseColors,typeColor,problem->getDataPtr());
    dlg.exec();
    if(dlg.isChanged()) {
        colorChanged=true;
        update();
    }
}

void CScene::on_actionSave_colour_scheme_triggered()
{
    QString file = QFileDialog::getSaveFileName(this,tr("File to save colours"),
                                                ".");

    if(file.isEmpty()) return;
    if(file.mid(file.lastIndexOf(".")+1).toLower()!="txt") file+=".txt";

    ofstream of(file.toStdString().c_str());
    if(!of) {
        QMessageBox::critical(this,"Error",tr("Could not create %1").arg(file));
        return;
    }

    size_t K=problem->getNumTypes();
    of << COLOR_END+K << endl;
    for(size_t i=0;i<COLOR_END;i++) {
        of << baseColors[i].red() << ','<< baseColors[i].green()<<','<<baseColors[i].blue()<<','<<baseColors[i].alpha()<<endl;
    }
    for(size_t i=0;i<K;i++) {
        of <<typeColor[i].red() << ','<< typeColor[i].green()<<','<<typeColor[i].blue()<<','<<typeColor[i].alpha()<<endl;
    }
    of.close();
}

void CScene::on_actionLoad_colour_scheme_triggered()
{
    //get the file name
    QString file=QFileDialog::getOpenFileName(this,tr("Select file"), ".", tr("Colour files (*.txt)"));
    if(!file.size()) return;

    ifstream in(file.toStdString().c_str());
    if(!in) {
        QMessageBox::critical(NULL,tr("error"),tr("Could not open %1").arg(file));
        return;
    }
    //read the number of colours
    int n;
    int num_bc=COLOR_END;
    size_t line=1;
    vector<string> tk;
    in >> n;
    for(int i=0;i<(num_bc<=n?num_bc:n);i++) {
        line++;
        if(!readLine(in,tk,",",line)) {
            QMessageBox::critical(NULL,tr("error"),tr("Unexpected end of file at line %1").arg(line));
            in.close();
            return;
        }
        if(tk.size()!=4) {
            QMessageBox::critical(NULL,tr("error"),tr("Error reading colour at line %1").arg(line));
            in.close();
            return;
        }
        QColor cl(atoi(tk[0].c_str()),atoi(tk[1].c_str()),atoi(tk[2].c_str()),atoi(tk[3].c_str()));
        baseColors[i]=cl;
    }
    for(int i=0;i<n-num_bc;i++) {
        line++;
        if(!readLine(in,tk,",",line)) {
            QMessageBox::critical(NULL,tr("error"),tr("Unexpected end of file at line %1").arg(line));
            in.close();
            return;
        }
        if(tk.size()!=4) {
            QMessageBox::critical(NULL,tr("error"),tr("Error reading colour at line %1").arg(line));
            in.close();
            return;
        }
        QColor cl(atoi(tk[0].c_str()),atoi(tk[1].c_str()),atoi(tk[2].c_str()),atoi(tk[3].c_str()));
        typeColor[i]=cl;
    }
    colorChanged=true;
    update();
}
