/*
 * cthreads.cpp
 * Copyright (C) Raul Massano Brás 2012 <rbras@iseg.utl.pt>
 *
 * MulTyLink is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MulTyLink is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "cthreads.h"
#include "cproblem.h"
#include "cdata.h"
#include <algorithm>
#include <map>
#include <time.h>

#include <iostream>
using namespace std;

double bestCost;
graphSol_threads res;

QMutex mutexData;
my_msgbuf *msgBuf;
bool reportProgress=true;
int countProgress=0;
double totIter=0;


void CWorkingThread::run()
{

    int bufIndex;
    size_t K=problem->getNumTypes();
    CData *data=problem->getDataPtr();
    //const CData *data=problem->getDataPtr();
    //const params *pars=problem->getParams();

    for(bufIndex=begin;bufIndex<=end;bufIndex++) {

        if(msgBuf[bufIndex].mtype==THREAD_RUN1) {
            logMsg_s lm;
            vector<int> *permut=(vector<int> *)msgBuf[bufIndex].args.arg1;

            set<size_t> vertInSol; //save the ids of the vertices in solution
            graphSol_threads *currSol;
            currSol=new graphSol_threads [K];

            //Solve the problem starting with this permutaion of the indices
            for(size_t ii=0;ii<K;ii++) {
                int i=permut->at(ii);

                //Can't use the global gk[k] in simulateneous threads. Build a graph from
                //gk[k] and use it

                //build the graph for pa type i. rg contains descriptors of gk[i]
                CGraph_d *rg=buildDirectedGraphOfType(i,vertInSol);

                //Calculate the shortest path between all pairs of reserves of this type
                //create a new graph
                CGraph_u *gspst=new CGraph_u;
                //insert the vertices. Ids are descriptors of gk[i]
                for(size_t j=0;j<terminals[i].size();j++) //for all terminals of type i
                    gspst->insertVertex((size_t)terminals[i][j]); //descriptors in gk[i]

                //data structure to store the shortest path between any 2 reserves

                map< pair<vDesc_u,vDesc_u>, caminho_d_t > paths;


                //We have to calculate paths starting at 0,...,terminais[i].size()-2 to all the others


                for(int j=0;j<(int)terminals[i].size()-1;++j) {
                    vector<vDesc_d> dest;
                    for(size_t jj=j+1;jj<terminals[i].size();++jj) {
                        dest.push_back((vDesc_d)(rg->findVertex((size_t)terminals[i][jj]).first));
                    }
                    if(!dest.size()) continue;

                    vector<caminho_d_t> path;
                    rg->dijkstra((vDesc_d)(rg->findVertex((size_t)terminals[i][j]).first),dest,path);

                    vDesc_u u_gspst=gspst->findVertex((size_t)terminals[i][j]).first;
                    vDesc_u u=terminals[i][j];

                    for(size_t jj=0;jj<path.size();++jj) {
                        if(!path[jj].l.size()) continue;
                        //convert descriptors in the path to descriptors of gk[i] (ids in rg are descriptors in gk[i])
                        for(list<vDesc_d>::iterator it=path[jj].l.begin();it!=path[jj].l.end();++it) {
                            *it=(vDesc_u)rg->getId(*it);
                        }
                        //Insert the edges between the terminals in gspst
                        vDesc_u v_gspst=gspst->findVertex((size_t)path[jj].l.back()).first;
                        gspst->insertEdge(u_gspst,v_gspst,path[jj].cost);
                        //Save the path. The lower index is the first in the pair
                        vDesc_u v=path[jj].l.back();
                        if(u>v) swap(u,v);
                        paths[make_pair(u,v)]=path[jj];
                    }
                }


                //Calculate connected components in gspst
                vector< vector<vDesc_u> > conn;
                gspst->connComp(conn);

                //In each connected component calculate the MST

                set<vDesc_u> solAux;

                for(size_t j=0;j<conn.size();j++) {
                    vector< pair<vDesc_u,vDesc_u> > mst;
                    gspst->prim(conn[j][0],mst);
                    //save the original vertices from wich we built the edges of gspst in the
                    //temporary solution
                    //we have pairs of terminals in mst
                    for(size_t jj=0;jj<mst.size();jj++) {
                        vDesc_u u=(vDesc_u)gspst->getId(mst[jj].first); // descriptor in gk[i]
                        vDesc_u v=(vDesc_u)gspst->getId(mst[jj].second); // descriptor in gk[i]
                        if(u>v) swap(u,v);
                        solAux.insert(paths[make_pair(u,v)].l.begin(),paths[make_pair(u,v)].l.end());
                    }
                }

                delete rg;
                delete gspst;
                paths.clear();

                //We have a solution in solAux but it contains cycles. We must eliminate them
                cleanCyclesFromType(solAux,i);

                //Now solAux contains the solution for type i (descriptors of gk[i])
                //add this solution to the set of vertices in solution. vertInSol contains Ids

                for(set<vDesc_u>::iterator it=solAux.begin();it!=solAux.end();++it) {
                    vertInSol.insert(gk[i]->getId(*it));
                }

                //calculate the cost of this solution for type i
                double cost=0;
                for(set<vDesc_u>::iterator it=solAux.begin();it!=solAux.end();it++) {
                    cost+=data->getCost(gk[i]->getId(*it));
                }

                //now save it.
                currSol[i].vertices=solAux;
                currSol[i].cost=cost;

                mutexData.lock();
                if(reportProgress) {
                    countProgress++;
                    emit threadProgressSignal((int)(0.5+100.0*countProgress/(K*totIter)));
                }
                mutexData.unlock();

            }


            //Start by joining all sites eliminating duplicates
            //Convert descriptors to ids


            set<vDesc_u> allSites;

            for(size_t k=0;k<K;k++) {
                for(set<vDesc_u>::iterator it=currSol[k].vertices.begin();it!=currSol[k].vertices.end();++it) {
                    allSites.insert((vDesc_u)gk[k]->getId(*it));
                }
            }


            //Make it minmal. currSol[i] contains decriptors of gk[i], allSites contains ids
            CGraph_u **gsol=makeMinimal(currSol,allSites);

            //Now calculate the cost
            double c=0;
            for(set<vDesc_u>::iterator it=allSites.begin();it!=allSites.end();it++) {
                c+=data->getCost((size_t)*it);
            }

            //bestCost and res are global. Protect access to them

            mutexData.lock();
            //lm.msg=tr("Permutation: %1").arg(1+permut->at(0));
            //for(size_t k=1;k<K;k++) lm.msg+=tr(",%1").arg(1+permut->at(k));
            //emit threadMsgSignal(MSG_LOG,0,lm);

            //lm.msg=tr("Cost: %1").arg(c);
            //emit threadMsgSignal(MSG_LOG,0,lm);

            //cout << "Cost: " << c << " (N. sites: " << allSites.size() << ')'<<endl;
            //If it is the best so far, save it
            if(bestCost>c) {
                res.vertices=allSites;
                res.cost=c;
                bestCost=c;
                if(res.gsol) {
                    for(size_t k=0;k<K;k++) delete res.gsol[k];
                    delete [] res.gsol;
                }
                res.gsol=gsol;
                lm.msg=tr("Best cost: %1\n").arg(bestCost);
                emit threadMsgSignal(MSG_LOG,0,lm);
                //cout <<"Best cost: " << bestCost << endl;
            }
            else {
                for(size_t k=0;k<K;k++) delete gsol[k];
                delete [] gsol;
            }
            mutexData.unlock();
            delete [] currSol;
        }
        else if(msgBuf[bufIndex].mtype==THREAD_RUN2) {
            //Calculate the solution for each type from the global solution
            //for type k we want the vertices and the cost
            size_t k=(size_t)msgBuf[bufIndex].args.arg1;
            problemSol *pSol=(problemSol *)msgBuf[bufIndex].args.arg2;

            CGraph_d *rg=buildDirectedGraphOfTypeFromSol(k);

            //rg contains ids of sites and terminals

            //To find the solution for each PA type
            //calculate the shortest path between all pairs of reserves of this type
            //We have to calculate paths starting at 0,...,terminals[i].size()-2 to all the others

            set<size_t> thisSol;
            for(int j=0;j<(int)terminals[k].size()-1;++j) {

                bool exists;
                vDesc_d ori;
                size_t id=gk[k]->getId(terminals[k][j]);
                tie(ori,exists)=rg->findVertex(id);
                if(!exists)continue;

                vector<vDesc_d> dest;
                for(size_t jj=j+1;jj<terminals[k].size();++jj) {
                    vDesc_d v;
                    id=gk[k]->getId(terminals[k][jj]);
                    tie(v,exists)=rg->findVertex(id);
                    if(!exists) continue;
                    dest.push_back(v);
                }
                if(!dest.size()) continue;
                vector<caminho_d_t> path;
                rg->dijkstra(ori,dest,path);

                //Save solution
                for(size_t jj=0;jj<path.size();++jj) {
                    if(!path[jj].l.size()) continue;
                    //convert descriptots in the path to ids
                    for(list<vDesc_d>::iterator it=path[jj].l.begin();it!=path[jj].l.end();++it) {
                        thisSol.insert(rg->getId(*it));
                    }
                }
            }
            delete rg;

            //Calculate cost and save solution (terminals are ignored)
            double c=0;
            for(set<size_t>::iterator it=thisSol.begin();it!=thisSol.end();++it) {                                                                
                const vertex_data *vd=data->getVertexDataById(*it);
                if(vd->siteType==K+1) { //Multiple types
                    if(!binary_search(vd->cellTypes.begin(),vd->cellTypes.end(),k+1)) {
                        c+=data->getCost(*it);
                        pSol->sol[k].vertices.push_back(*it);
                    }
                }
                else if(vd->siteType!=(int)k+1) {
                    c+=data->getCost(*it);
                    pSol->sol[k].vertices.push_back(*it);
                }
            }
            pSol->sol[k].cost=c;


        }
    }
    return;
}

CGraph_d *CWorkingThread::buildDirectedGraphOfType(size_t k,set<size_t>& vertInSol)
{
    CGraph_d *rg=new CGraph_d;
    double epsilon=problem->getParams()->epsilon;
    CData *data=problem->getDataPtr();

    //Build a directed graph from gk[k]. Ids are descriptors of gk[k]

    //Insert vertices
    vIter_u vi,vi_end;
    for(tie(vi,vi_end)=gk[k]->getVertices();vi!=vi_end;++vi) {
        rg->insertVertex((size_t)*vi); //Ids descriptors of gk[k]
    }

    //Insert Edges
    eIter_u ei,ei_end;
    bool flag;
    for(tie(ei,ei_end)=gk[k]->getEdges();ei!=ei_end;++ei) {
        vDesc_u ori_gk=gk[k]->getEdgeSource(*ei);
        vDesc_u dest_gk=gk[k]->getEdgeTarget(*ei);

        vDesc_d ori,dest;
        tie(ori,flag)=rg->findVertex((size_t)ori_gk);
        tie(dest,flag)=rg->findVertex((size_t)dest_gk);

        rg->insertEdge(ori,dest,0);
        rg->insertEdge(dest,ori,0);

    }

    //Insert the costs. The cost is:
    // 0  - if the target is a terminal of type k
    // epsilon - if target is in solution
    // vertex cost  +  epsilon otherwise
    eIter_d ei1,ei1_end;
    for(tie(ei1,ei1_end)=rg->getEdges();ei1!=ei1_end;++ei1) {

        //Start by obtaining the Ids in rg
        vDesc_d desc_v_rg=rg->getEdgeTarget(*ei1);
        //get the id
        vDesc_u desc_gk=(vDesc_u)(rg->getId(desc_v_rg)); //descriptor in gk[k]
        size_t id=gk[k]->getId(desc_gk);
        double c=0.0;
        const vertex_data *vd=data->getVertexDataById(id);


        if(vd->siteType==(int)k+1 || (vd->siteType==data->getNumTypes()+1 && binary_search(vd->cellTypes.begin(),vd->cellTypes.end(),k+1) )) c=0.0; //terminal of type k
        else if(vertInSol.find(id)==vertInSol.end()) c=data->getCost(id)+epsilon; //Not in solution
        else c=0;
        rg->setEdgeCost(*ei1,c);
    }


    return rg;

}



CGraph_d *CWorkingThread::buildDirectedGraphOfTypeFromSol(size_t k)
{
    CGraph_d *ngk=new CGraph_d;
    CData *data=problem->getDataPtr();

    //The solution is in the graph res.sol[k]

    //Build a directed graph

    //Insert vertices
    vIter_u vi,vi_end;
    for(tie(vi,vi_end)=res.gsol[k]->getVertices();vi!=vi_end;++vi) {
        size_t id=res.gsol[k]->getId(*vi);
        ngk->insertVertex(id);
    }

    //Insert the arcs.
    eIter_u ei,ei_end;
    for(tie(ei,ei_end)=res.gsol[k]->getEdges();ei!=ei_end;++ei) {

        //Start by obtaining the Ids in gk
        size_t id_u,id_v;
        id_u=res.gsol[k]->getId( res.gsol[k]->getEdgeSource(*ei) );
        id_v=res.gsol[k]->getId( res.gsol[k]->getEdgeTarget(*ei) );
        vDesc_d idx_u_ngk,idx_v_ngk;
        bool flag;
        tie(idx_u_ngk,flag)=ngk->findVertex(id_u);
        tie(idx_v_ngk,flag)=ngk->findVertex(id_v);
        //Insert arc u-->v
        //Calculate the cost
        double c=data->getCost(id_v);
        ngk->insertEdge(idx_u_ngk,idx_v_ngk,c);
        //Insert arc v-->u
        //Calculate the cost
        c=data->getCost(id_u);
        ngk->insertEdge(idx_v_ngk,idx_u_ngk,c);
    }

    return ngk;

}



void CWorkingThread::cleanCyclesFromType(set<vDesc_u>& sol,size_t type)
{
    //Build a graph from sol.
    //sol contains descriptors in gk[type]

    CData *data=problem->getDataPtr();
    CGraph_u *cg=new CGraph_u;

    //Insert vertices
    for(set<vDesc_u>::iterator it=sol.begin();it!=sol.end();++it) {
        cg->insertVertex((size_t)*it);
    }


    //Insert adjacencies
    vIter_u vi,vi_end;
    for(tie(vi,vi_end)=cg->getVertices();vi!=vi_end;++vi) {
        vDesc_u idx=(vDesc_u)cg->getId(*vi); //The Id in cg is a descriptor in gk[type]
        //For all original adjacencies
        aIter_u ai,ai_end;
        for(tie(ai,ai_end)=gk[type]->getAdjVertices(idx);ai!=ai_end;++ai) {
            bool flag;
            vDesc_u v;
            tie(v,flag)=cg->findVertex((size_t)*ai);
            if(!flag) continue;
            cg->insertEdge(*vi,v,0.0);
        }
    }

    //Calculate connected components
    vector< vector<vDesc_u> > comps;
    cg->connComp(comps);

    vector<vDesc_u> sAdj;
    size_t numComps=comps.size();


    for(set<vDesc_u>::iterator it=sol.begin();it!=sol.end();++it) {
        size_t id=gk[type]->getId(*it);

        const vertex_data *vd=data->getVertexDataById(id);
        if(vd->siteType==(int)type+1 || (vd->siteType==data->getNumTypes()+1 && binary_search(vd->cellTypes.begin(),vd->cellTypes.end(),type+1) )) continue;
        vDesc_u idx=cg->findVertex((size_t)*it).first;
        //save the adjacency list of vertex before removing
        aIter_u ai,ai_end;
        sAdj.clear();
        for(tie(ai,ai_end)=cg->getAdjVertices(idx);ai!=ai_end;++ai) sAdj.push_back(*ai);
        //Finnished saving. Remove it
        cg->removeVertex(idx);
        //See if any component was broken
        comps.clear();
        cg->connComp(comps);
        if(comps.size()>numComps) { //Can't remove it
            vDesc_u v=cg->insertVertex((size_t)*it);
            for(size_t j=0;j<sAdj.size();j++) cg->insertEdge(v,sAdj[j],0);
        }
    }

    //rebuild the solution
    sol.clear();
    for(tie(vi,vi_end)=cg->getVertices();vi!=vi_end;++vi) {
        sol.insert((vDesc_u)(cg->getId(*vi)));
    }

    delete cg;
}

CGraph_u **CWorkingThread::makeMinimal(graphSol_threads *currSol, set<vDesc_u>& sol)
{
    //currSol[k] contains descriptors of gk[i]
    //sol contains ids

    CData *data=problem->getDataPtr();
    size_t K=problem->getNumTypes();
    //Start by building a graph from solution. The graphs ngk conatain ids

    CGraph_u **ngk;
    ngk=new CGraph_u * [K];
    for(size_t i=0;i<K;i++) ngk[i]=new CGraph_u;

    for(size_t k=0;k<K;k++) {
        //Insert the vertices in the solution for type k
        for(set<vDesc_u>::iterator it=currSol[k].vertices.begin();it!=currSol[k].vertices.end();++it) {
            ngk[k]->insertVertex(gk[k]->getId(*it)); //ids of sites and terminals
        }
        //Insert the vertices in the solution for other types that are also in gk[k]
        for(size_t i=0;i<K;i++) {
            if(i==k) continue;
            bool flag;
            vDesc_u v;
            for(set<vDesc_u>::iterator it=currSol[i].vertices.begin();it!=currSol[i].vertices.end();++it) {
                size_t id=gk[i]->getId(*it);
                tie(v,flag)=gk[k]->findVertex(id);
                if(flag) ngk[k]->insertVertex(id);
            }
        }
    }

    //Insert edges
    for(size_t k=0;k<K;k++) {
        //Insert the original edges if source and target exist em ngk
        eIter_u ei,ei_end;
        bool flag;
        vDesc_u ori,dest;
        for(tie(ei,ei_end)=gk[k]->getEdges();ei!=ei_end;++ei) {
            vDesc_u u=gk[k]->getEdgeSource(*ei);
            tie(ori,flag)=ngk[k]->findVertex(gk[k]->getId(u));
            if(!flag) continue;
            vDesc_u v=gk[k]->getEdgeTarget(*ei);
            tie(dest,flag)=ngk[k]->findVertex(gk[k]->getId(v));
            if(!flag) continue;
            ngk[k]->insertEdge(ori,dest,0);
        }
    }

    for(size_t k=0;k<K;k++) {
        vector< vector<vDesc_u> > vconn;
        ngk[k]->connComp(vconn);
        for(size_t i=0;i<vconn.size();i++) {
            bool hasTerm=false;
            for(size_t j=0;j<vconn[i].size();j++) {
                size_t id=ngk[k]->getId(vconn[i][j]);
                const vertex_data *vd=data->getVertexDataById(id);
                if(vd->siteType==(int)k+1 || (vd->siteType==data->getNumTypes()+1 && binary_search(vd->cellTypes.begin(),vd->cellTypes.end(),k+1) )) {
                    hasTerm=true; break;
                }
            }
            if(!hasTerm) { //Remove the component from the graph
                for(size_t j=0;j<vconn[i].size();j++) {
                    ngk[k]->removeVertex(vconn[i][j]);
                }
            }
        }
    }


    //Now try to remove vertices from the graphs


    vector< set<vDesc_u> >comps_s[K];
    //size_t nr=0;
    for(set<vDesc_u>::iterator it=sol.begin();it!=sol.end();++it) {
        if(data->isTerm((size_t) *it)) continue; //do not try to remove terminals

        //bool remover=true;
        vector<vDesc_u> removidos_adj[K];
        vector<size_t> retirados;
        for(size_t k=0;k<K;k++) {
            comps_s[k].clear();
            //if the vertex is not in ngk continue

            vDesc_u idx_gk;
            bool flag;
            tie(idx_gk,flag)=ngk[k]->findVertex((size_t) *it);
            if(!flag) continue;

            if(!removeVertex(ngk[k],idx_gk,k,comps_s[k],removidos_adj[k])) {
                //remover=false;
                //put back the vertex in ngk[k]
                vDesc_u v=ngk[k]->insertVertex((size_t)*it);
                for(size_t ii=0;ii<removidos_adj[k].size();ii++) {
                    ngk[k]->insertEdge(v,removidos_adj[k][ii],0);
                }
                //put back the ones we have removed before
                for(size_t j=0;j<retirados.size();j++) {
                    //Vertex
                    vDesc_u v=ngk[retirados[j]]->insertVertex((size_t) *it);
                    //Adjacencies
                    for(size_t ii=0;ii<removidos_adj[retirados[j]].size();ii++) {
                        ngk[retirados[j]]->insertEdge(v,removidos_adj[retirados[j]][ii],0);
                    }
                }
                break;
            }
            else retirados.push_back(k);
        }
    }



    //update the solution

    sol.clear();
    vIter_u vi,vi_end;
    for(size_t k=0;k<K;k++) {
        for(tie(vi,vi_end)=ngk[k]->getVertices();vi!=vi_end;++vi) sol.insert((vDesc_u) ngk[k]->getId(*vi));
    }


    return ngk;

}


bool CWorkingThread::removeVertex(CGraph_u *ngk,vDesc_u idx,size_t k,vector< set<vDesc_u> >& comps,
                                  vector<vDesc_u>& adjs)
{
    //idx is the index in gk
    //before removing the vertex save adjacencies

    aIter_u ai,ai_end;
    tie(ai,ai_end)=ngk->getAdjVertices(idx);
    while(ai!=ai_end) {
        adjs.push_back(*ai);
        ai++;
    }

    ngk->removeVertex(idx);

    //Calculate connected components

    vector< vector<vDesc_u> > vconn;
    ngk->connComp(vconn);
    //vconn contains descriptors of gk.

    for(size_t i=0;i<vconn.size();i++) {
        set<vDesc_u> x;
        comps.push_back(x);
        for(size_t j=0;j<vconn[i].size();j++) {
            comps[i].insert(gk[k]->findVertex(ngk->getId(vconn[i][j])).first); //change to descriptors of gk[k]
        }
    }

    //If anny connected component is violated we can't remove the vertex
    for(size_t i=0;i<comps.size();i++) {
        if(fviol(comps[i],k)) return false;
    }

    return true;
}

bool CWorkingThread::fviol(const set<vDesc_u>& conn,size_t k)
{
    //Check if any component is violated. A component is violated if for
    //any pair of terminals it contains one but not the other
    //reurns: true if the component is violated, false otherwise

    bool existe1,existe2;
    for(set< pair<vDesc_u,vDesc_u> >::iterator it=term_par[k].begin();it!=term_par[k].end();it++) {
        existe1=false; existe2=false;
        if(conn.find(it->first)!=conn.end()) existe1=true;
        if(conn.find(it->second)!=conn.end()) existe2=true;
        if(existe1!=existe2) return true;
    }

    return false;
}


void CSolveThread::run()
{

    logMsg_s lm;
    CData *data=problem->getDataPtr();
    const params *pars=problem->getParams();

    //fill the terminals vector. It contains descriptors of
    //the gk graphs

    for(size_t k=0;k<K;k++) {
        vIter_u vi,vi_end;
        for(tie(vi,vi_end)=gk[k]->getVertices();vi!=vi_end;++vi) {
            if(data->getSiteType(gk[k]->getId(*vi))==(int)k+1) terminals[k].push_back(*vi);
        }
        sort(terminals[k].begin(),terminals[k].end());
    }

    //Verify the graph and initialize term_par and rep. We also calculate
    //the number of components for each type

    lm.msg=tr("Checking graph:");
    emit threadMsgSignal(MSG_LABEL,0,lm);
    countProgress=0;

    verifyGraph(); //Initializes term_par and rep


    //cout << "Numero de terminais:"<<endl;
    //for(size_t i=0;i<K;i++) cout << "\tTipo "<<i+1<<": "<<terminals[i].size()<<endl;

    vector< vector<int> > exec_permut;

    if(K<=4) {
        //Calculate the number of possible permutations
        size_t numPermut=1;
        for(size_t k=2;k<=K;k++) numPermut*=k;

        //The user may have spcified less then numPerm permutations to execute. If so
        //choose wich ones to deal with

        vector<int> possiblePerms(numPermut);
        for(size_t i=0;i<numPermut;i++) possiblePerms[i]=i+1;

        //Remove the excess of permutations

        int j=numPermut-1;
        for(int i=0;i<(int)(numPermut-pars->nperm);i++) {
            int idx=randInt(0,j);
            possiblePerms.erase(possiblePerms.begin()+idx);
            j--;
        }
        j=0;
        vector<int> permut(K);
        for(size_t i=0;i<K;i++) permut[i]=i;
        do{
            j++;
            if(!binary_search(possiblePerms.begin(),possiblePerms.end(),j)) continue;
            exec_permut.push_back(permut);
        }while(next_permutation(permut.begin(),permut.end()));

    }
    else {
        for(int i=0;i<(int)pars->nperm;i++) {
            vector<int> v;
            shuffle(K,v);
            bool repeated=false;
            for(size_t j=0;j<exec_permut.size();++j) {
                if(exec_permut[j]==v) {repeated=true; break;}
            }
            if(repeated) {i--; continue;}
            exec_permut.push_back(v);
        }
    }



    bestCost=1e+100;
    res.gsol=NULL;

    //Send each permutation to the threads
    totIter=pars->nperm;

    msgBuf=new my_msgbuf [max(K,pars->nperm)];
    size_t index=0;

    for(size_t i=0;i<exec_permut.size();++i) {
        msgBuf[index].mtype=THREAD_RUN1;
        msgBuf[index].args.arg1=(void *)&exec_permut[index];
        index++;
    }

    /////////////////

    lm.msg=tr("Calculating solution:");
    emit threadMsgSignal(MSG_LABEL,0,lm);
    countProgress=0;

    //Create the threads
    int num_t=min(nt,index);
    int begin=0,end;
    for(int t=0;t<num_t;t++) {
        if(t==num_t-1) end=index-1;
        else end=begin+index/num_t-1;
        thread[t]=new CWorkingThread(gk,problem,terminals,rep,term_par,begin,end);
        connect(thread[t],SIGNAL(threadProgressSignal(int)),scene,SLOT(threadProgressMessages(int)));
        connect(thread[t],SIGNAL(threadMsgSignal(int, uint, logMsg_s )),scene,SLOT(threadMessages(int, uint, logMsg_s)));
        thread[t]->start();
        begin=end+1;
    }

    //Wait for threads to finish
    for(int t=0;t<num_t;t++) {
        if(thread[t]) {
            thread[t]->wait();
            mutexData.lock();
            disconnect(thread[t],0,0,0);
            delete thread[t];
            thread[t]=NULL;
            mutexData.unlock();
        }
    }


    ////////////////


    //cout<<"ended fase 1"<<endl;

    //Find solution for each type an calculate the cost


    for(size_t i=0;i<K;i++) {
        msgBuf[i].mtype=THREAD_RUN2;
        msgBuf[i].args.arg1=(void *)i;
        msgBuf[i].args.arg2=(void *)resSol;
    }

    /////////////////////////

    countProgress=0;
    totIter=K;

    //Create the threads
    num_t=min(nt,K);
    begin=0;
    for(int t=0;t<num_t;t++) {
        if(t==num_t-1) end=K-1;
        else end=begin+K/num_t-1;
        thread[t]=new CWorkingThread(gk,problem,terminals,rep,term_par,begin,end);
        connect(thread[t],SIGNAL(threadProgressSignal(int)),scene,SLOT(threadProgressMessages(int)));
        connect(thread[t],SIGNAL(threadMsgSignal(int, uint, logMsg_s )),scene,SLOT(threadMessages(int, uint, logMsg_s)));
        thread[t]->start();
        begin=end+1;
    }

    //Wait for threads to finish
    for(int t=0;t<num_t;t++) {
        if(thread[t]) {
            thread[t]->wait();
            mutexData.lock();
            disconnect(thread[t],0,0,0);
            delete thread[t];
            thread[t]=NULL;
            mutexData.unlock();
        }
    }

    /////////////////////////


    //cout<<"ended fase 2"<<endl;

    //cout << endl << "Processing finnished. Emiting solution" << endl;

    //Output the solution

    lm.msg=tr("\nSolution with friction threshold=%1\n").
            arg(pars->limit);
    emit threadMsgSignal(MSG_LOG,0,lm);

    for(size_t i=0;i<K;i++) {
        lm.msg=tr("\nSolution for %1").arg(QString(data->getPAName(i).c_str()));
        emit threadMsgSignal(MSG_LOG,0,lm);;
        for(size_t j=0;j<resSol->sol[i].vertices.size();j++) {
            //const vertex_data *vd=getVertexDataById(res[i].vertices[j]);
            lm.msg=tr("%1").arg(resSol->sol[i].vertices[j]);
            emit threadMsgSignal(MSG_LOG,0,lm);
        }
    }

    //Calculate cost by type

    for(size_t i=0;i<K;i++) {
        resSol->max_friction[i]=resSol->max_cost[i]=0;
        resSol->min_friction[i]=resSol->min_cost[i]=1e+30;
    }
    for(size_t i=0;i<K;i++) {
        lm.msg=trUtf8("\nType %1:").arg(QString(data->getPAName(i).c_str()));
        emit threadMsgSignal(MSG_LOG,0,lm);
        double cc=0,ss=0;
        for(size_t j=0;j<resSol->sol[i].vertices.size();j++) {
            const vertex_data *vd=data->getVertexDataById(resSol->sol[i].vertices[j]);
            if(vd->siteType==0) ss++;
            cc+=vd->cost;
            if(resSol->max_friction[i]<vd->friction) resSol->max_friction[i]=vd->friction;
            if(resSol->min_friction[i]>vd->friction) resSol->min_friction[i]=vd->friction;
            if(resSol->max_cost[i]<vd->cost) resSol->max_cost[i]=vd->cost;
            if(resSol->min_cost[i]>vd->cost) resSol->min_cost[i]=vd->cost;
        }
        resSol->cost[i]=cc;

        resSol->nSites[i]=ss;
        lm.msg=trUtf8("\tCost:                  %1").arg(cc);
        emit threadMsgSignal(MSG_LOG,0,lm);
        lm.msg=trUtf8("\tMin. friction:               %1").arg(resSol->min_friction[i]);
        emit threadMsgSignal(MSG_LOG,0,lm);
        lm.msg=trUtf8("\tMax. friction:               %1").arg(resSol->max_friction[i]);
        emit threadMsgSignal(MSG_LOG,0,lm);
        lm.msg=trUtf8("\tMin. cost:           %1").arg(resSol->min_cost[i]);
        emit threadMsgSignal(MSG_LOG,0,lm);
        lm.msg=trUtf8("\tMax. cost:           %1").arg(resSol->max_cost[i]);
        emit threadMsgSignal(MSG_LOG,0,lm);
        lm.msg=trUtf8("\tNum. Comps :            %1").arg(resSol->numComps[i]);
        emit threadMsgSignal(MSG_LOG,0,lm);
    }
    lm.msg=trUtf8("\nTotals:");
    emit threadMsgSignal(MSG_LOG,0,lm);

    //Calculate total cost

    //Start by joining all sites eliminating duplicates
    set<int> tmp,allSites;
    for(size_t i=0;i<K;i++) tmp.insert(resSol->sol[i].vertices.begin(),resSol->sol[i].vertices.end());
    //Eliminate terminals
    for(set<int>::iterator it=tmp.begin();it!=tmp.end();++it)
        if(data->getSiteType(*it)==0) allSites.insert(*it);

    tmp.clear();



    //Now calculate it
    double cc=0;
    resSol->totSites=0;
    for(set<int>::iterator it=allSites.begin();it!=allSites.end();it++) {
        const vertex_data *vd=data->getVertexDataById(*it);
        if(vd->siteType==0) resSol->totSites++; //Count only the sites not belonging to PA
        cc+=vd->cost;
    }

    resSol->totCost=cc;
    resSol->totSites=allSites.size();
    lm.msg=trUtf8("\tCost:           %1").arg(resSol->totCost);
    emit threadMsgSignal(MSG_LOG,0,lm);
    lm.msg=trUtf8("\tNum. sites:     %1").arg(resSol->totSites);
    emit threadMsgSignal(MSG_LOG,0,lm);
    lm.msg=trUtf8("\tMin. friction: %1").arg(*min_element(resSol->min_friction,resSol->min_friction+K));
    emit threadMsgSignal(MSG_LOG,0,lm);
    lm.msg=trUtf8("\tMax. friction: %1").arg(*max_element(resSol->max_friction,resSol->max_friction+K));
    emit threadMsgSignal(MSG_LOG,0,lm);
    lm.msg=trUtf8("\tMin. cost:      %1").arg(*min_element(resSol->min_cost,resSol->min_cost+K));
    emit threadMsgSignal(MSG_LOG,0,lm);
    lm.msg=trUtf8("\tMax. cost:       %1").arg(*max_element(resSol->max_cost,resSol->max_cost+K));
    emit threadMsgSignal(MSG_LOG,0,lm);
    lm.msg="";

    delete [] msgBuf;
    if(res.gsol) {
        for(size_t i=0;i<K;i++) delete res.gsol[i];
        delete [] res.gsol;
    }

    emit threadMsgSignal(MSG_FINISHED,0,lm);

    //cout << "exiting CSolveThread" << endl;

}



bool CSolveThread::verifyGraph()
{

    //For each type verify if terminals are in the same component
    //Components contain descriptors of gk[k] and are sorted

    for(size_t k=0;k<K;k++) {
        resSol->numComps[k]=0;
        vector< vector<vDesc_u> > vconn;
        gk[k]->connComp(vconn);

        vector<vDesc_u> *term_in_comp;
        term_in_comp=new vector<vDesc_u> [vconn.size()];

        for(size_t i=0;i<vconn.size();i++) {
            sort(vconn[i].begin(),vconn[i].end());

            for(size_t j=0;j<terminals[k].size();j++) {
                if(binary_search(vconn[i].begin(),vconn[i].end(),terminals[k][j])) {
                    term_in_comp[i].push_back(terminals[k][j]);
                }
            }
        }
        for(size_t i=0;i<vconn.size();i++) {
            if(term_in_comp[i].size()!=0) {
                resSol->numComps[k]++;
                //Insert terminal pairs
                if(term_in_comp[i].size()>1) {
                    //save a representative of the terminals in the component
                    rep[k].push_back(term_in_comp[i][0]);
                    for(size_t j=1;j<term_in_comp[i].size();j++) {
                        term_par[k].insert(make_pair(term_in_comp[i][0],term_in_comp[i][j]));
                    }
                }
                else {
                    //The terminal is isolated. Discard it form terminals[k]
                    vector<vDesc_u>::iterator it=find(terminals[k].begin(),terminals[k].end(),term_in_comp[i][0]);
                    terminals[k].erase(it);
                }
            }
        }
        delete [] term_in_comp;
    }


    return true;
}
void CSolveThread::shuffle(int n, vector<int>& v)
{
  //returns in v a random permutation of the values {0,1,2,...,n-1}

  //get space for n elements
  v.resize(n);

  //initial available positions
  vector<int> availablePos(n);
  for(int i=0;i<n;i++) availablePos[i]=i;

  //fill it
  for(int i=0;i<n;i++) {
    int idx=randInt(0,availablePos.size()-1);
    v[availablePos[idx]]=i;
    availablePos.erase(availablePos.begin()+idx);
  }
}
