/*
 * cscene.cpp
 * Copyright (C) Raul Massano Brás 2012 <rbras@iseg.utl.pt>
 *
 * MulTyLink is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MulTyLink is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "cthreadsgreedy.h"


void CThreadsGreedy::run()
{
    //Qt uses a thread safe random generator (which one?). The random number
    //sequence is local to the threads. It would be a better idea to have a
    //global generator based on Mersenne twister or similar and initialize it
    //only once when MulTyLink starts. To be done later.

    QTime time_r = QTime::currentTime();
    qsrand((uint)time_r.msec());

    QTime time;
    time.start();
    logMsg_s lm;


    initTermPar(); //Initialize terms_in_comp and term_par

    // << "Numero de terminais:"<<endl;
    //for(int i=0;i<K;i++) cout << "\tTipo "<<i+1<<": "<<terminals[i].size()<<endl;


    set<int> bestSol;
    double bestCost=1e+100;

    int count=0;
    for(;;) {
        set<int> sol;
        calcSolution(sol);
        makeMinimal(sol);

        //cost of solution
        double cost=0.0;
        for(set<int>::iterator it=sol.begin();it!=sol.end();++it)
            cost+=data->getCost(*it);

        if(cost<bestCost) {
            bestCost=cost;
            bestSol=sol;
            //cout << "Melhor custo: " << bestCost << endl;
        }

        count++;
        emit threadProgressSignal((int)(0.5+100.0*count/(pars->numRep)));
        if(count==pars->numRep) break;

        if(time.elapsed()/1000.0>pars->time_limit) {
            lm.msg=tr("Algorithm ended by time limit. Number of runs: %1").
                    arg(count);
            emit threadMsgSignal(MSG_LOG,0,lm);
            break;
        }

        //Reset costs of gk for vertices in solution to original values
        for(int k=0;k<K;k++) {
            eIter_d ei,ei_end;
            for(tie(ei,ei_end)=gk[k].getEdges();ei!=ei_end;++ei){
                vDesc_d dest=gk[k].getEdgeTarget(*ei);
                int id=gk[k].getId(dest);
                double c=(data->getSiteType(id)==k+1?0:data->getCost(id)+pars->epsilon);
                gk[k].setEdgeCost(*ei,c);
            }
        }                                     

    }

    //Calculate solutions for each type. They are saved in resSol;
    calcSolutionByType(bestSol);

    //Output the solution

    lm.msg=tr("\nSolution with friction threshold=%1\n").
            arg(pars->limit);
    emit threadMsgSignal(MSG_LOG,0,lm);

    for(int i=0;i<K;i++) {
        lm.msg=tr("\nSolution for %1").arg(QString(data->getPAName(i).c_str()));
        emit threadMsgSignal(MSG_LOG,0,lm);;
        for(size_t j=0;j<resSol->sol[i].vertices.size();j++) {
            //const vertex_data *vd=getVertexDataById(res[i].vertices[j]);
            lm.msg=tr("%1").arg(resSol->sol[i].vertices[j]);
            emit threadMsgSignal(MSG_LOG,0,lm);
        }
    }

    //Calculate cost by type

    for(int i=0;i<K;i++) {
        resSol->max_friction[i]=resSol->max_cost[i]=0;
        resSol->min_friction[i]=resSol->min_cost[i]=1e+30;
    }
    for(int i=0;i<K;i++) {
        lm.msg=trUtf8("\nType %1:").arg(QString(data->getPAName(i).c_str()));
        emit threadMsgSignal(MSG_LOG,0,lm);
        double cc=0,ss=0;
        for(size_t j=0;j<resSol->sol[i].vertices.size();j++) {
            const vertex_data *vd=data->getVertexDataById(resSol->sol[i].vertices[j]);
            if(vd->siteType==0) ss++;
            cc+=vd->cost;
            if(resSol->max_friction[i]<vd->friction) resSol->max_friction[i]=vd->friction;
            if(resSol->min_friction[i]>vd->friction) resSol->min_friction[i]=vd->friction;
            if(resSol->max_cost[i]<vd->cost) resSol->max_cost[i]=vd->cost;
            if(resSol->min_cost[i]>vd->cost) resSol->min_cost[i]=vd->cost;
        }
        resSol->cost[i]=cc;

        resSol->nSites[i]=ss;
        lm.msg=trUtf8("\tCost:                  %1").arg(cc);
        emit threadMsgSignal(MSG_LOG,0,lm);
        lm.msg=trUtf8("\tMin. friction:               %1").arg(resSol->min_friction[i]);
        emit threadMsgSignal(MSG_LOG,0,lm);
        lm.msg=trUtf8("\tMax. friction:               %1").arg(resSol->max_friction[i]);
        emit threadMsgSignal(MSG_LOG,0,lm);
        lm.msg=trUtf8("\tMin. cost:           %1").arg(resSol->min_cost[i]);
        emit threadMsgSignal(MSG_LOG,0,lm);
        lm.msg=trUtf8("\tMax. cost:           %1").arg(resSol->max_cost[i]);
        emit threadMsgSignal(MSG_LOG,0,lm);
        lm.msg=trUtf8("\tNum. Comps :            %1").arg(resSol->numComps[i]);
        emit threadMsgSignal(MSG_LOG,0,lm);
    }
    lm.msg=trUtf8("\nTotals:");
    emit threadMsgSignal(MSG_LOG,0,lm);

    //Calculate total cost

    //Start by joining all sites eliminating duplicates
    set<int> tmp,allSites;
    for(int i=0;i<K;i++) tmp.insert(resSol->sol[i].vertices.begin(),resSol->sol[i].vertices.end());
    //Eliminate terminals
    for(set<int>::iterator it=tmp.begin();it!=tmp.end();++it)
        if(data->getSiteType(*it)==0) allSites.insert(*it);

    tmp.clear();




    //Now calculate it
    double cc=0;
    resSol->totSites=0;
    for(set<int>::iterator it=allSites.begin();it!=allSites.end();it++) {
        const vertex_data *vd=data->getVertexDataById(*it);
        if(vd->siteType==0) resSol->totSites++; //Count only the sites not belonging to PA
        cc+=vd->cost;
    }

    resSol->totCost=cc;
    resSol->totSites=allSites.size();
    lm.msg=trUtf8("\tCost:           %1").arg(resSol->totCost);
    emit threadMsgSignal(MSG_LOG,0,lm);
    lm.msg=trUtf8("\tNum. sites:     %1").arg(resSol->totSites);
    emit threadMsgSignal(MSG_LOG,0,lm);
    lm.msg=trUtf8("\tMin. friction: %1").arg(*min_element(resSol->min_friction,resSol->min_friction+K));
    emit threadMsgSignal(MSG_LOG,0,lm);
    lm.msg=trUtf8("\tMax. friction: %1").arg(*max_element(resSol->max_friction,resSol->max_friction+K));
    emit threadMsgSignal(MSG_LOG,0,lm);
    lm.msg=trUtf8("\tMin. cost:      %1").arg(*min_element(resSol->min_cost,resSol->min_cost+K));
    emit threadMsgSignal(MSG_LOG,0,lm);
    lm.msg=trUtf8("\tMax. cost:       %1").arg(*max_element(resSol->max_cost,resSol->max_cost+K));
    emit threadMsgSignal(MSG_LOG,0,lm);
    lm.msg="";


    emit threadMsgSignal(MSG_FINISHED,0,lm);

    //cout << "exiting CThreadsGreedy" << endl;



}

void CThreadsGreedy::initTermPar()
{
    vector< vector<vDesc_d> > vconn;

    for(int i=0;i<K;i++) {
        resSol->numComps[i]=0;
        gk[i].strongConnComp(vconn);
        for(size_t j=0;j<vconn.size();j++) {
            vector<int> terms;
            for(size_t jj=0;jj<vconn[j].size();jj++) {
                int id=gk[i].getId(vconn[j][jj]);
                if(binary_search(terminals[i].begin(),terminals[i].end(),id))
                    terms.push_back(id);
            }
            if(terms.size()) resSol->numComps[i]++;
            if(terms.size()==1) {
                //cout << "Terminal " << terms[0] << " está isolado. Ignorado" << endl;
                //Remove it from terminals
                vector<int>::iterator it=find(terminals[i].begin(),terminals[i].end(),terms[0]);
                terminals[i].erase(it);
            }
            if(terms.size()<=1) continue;
            //Aqui temos os terminais do tipo i que pertencem à componente j
            terms_in_comp[i].push_back(terms);
            for(size_t jj=1;jj<terms.size();jj++) term_par[i].insert(make_pair(terms[0],terms[jj]));
        }
    }

}

void CThreadsGreedy::calcSolution(set<int>& sol)
{
    vector< vector<vDesc_u> > vconn;

    vector< CGraph_u * > *graphs=new vector< CGraph_u * > [K];

    for(int i=0;i<K;i++) {
        for(size_t j=0;j<terms_in_comp[i].size();j++) {
            //Here we have terminals of type i belonging to component j
            //Create an undirected graph and insert the terminals
            CGraph_u *gu=new CGraph_u;
            for(size_t jj=0;jj<terms_in_comp[i][j].size();jj++)
                gu->insertVertex(terms_in_comp[i][j][jj]);
            graphs[i].push_back(gu);
        }
    }

    //choose a type randomly
    //for the choosen type chose a graph randomly (each graph is one original component)
    //choose a random component in the graph
    //connect one terminal from the component to the closet one among the terminals
    //in the other components


    //Available types
    vector<int> availableTypes;

    //Available graphs
    vector< vector<int> > availableGraphs(K);
    for(int i=0;i<(int)K;i++) {
        for(int j=0;j<(int)graphs[i].size();j++) availableGraphs[i].push_back(j);
        if(availableGraphs[i].size()) availableTypes.push_back(i);
    }


    for(;;) {
        //Choose a random type
        int rtype;
        int n=availableTypes.size();
        if(!n) break;
        else if(n==1) rtype=availableTypes[0];
        else rtype=availableTypes[randInt(0,n-1)];

        //Choose a random graph (components)
        int numGraph;
        n=availableGraphs[rtype].size();
        if(n==1) numGraph=availableGraphs[rtype][0];
        else numGraph=availableGraphs[rtype][randInt(0,n-1)];

        //Choose a random component of the graph
        graphs[rtype][numGraph]->connComp(vconn);
        n=vconn.size();
        int comp;
        if(n==1) comp=0;
        else comp=randInt(0,n-1);

        //Connect a terminal from the choosen component to the closest one in other components

        vDesc_d ori=gk[rtype].findVertex(graphs[rtype][numGraph]->getId(vconn[comp][0])).first;
        vector<vDesc_d> dests;
        for(int i=0;i<(int)vconn.size();i++) {
            if(i==comp) continue;
            for(int j=0;j<(int)vconn[i].size();j++) {
                dests.push_back(gk[rtype].findVertex(graphs[rtype][numGraph]->getId(vconn[i][j])).first);
            }
        }
        caminho_d_t c;
        vDesc_d v; //closest vertex
        sort(dests.begin(),dests.end());
        gk[rtype].dijkstra_first(ori,dests,c,v);

        //connect ori to dest in graphs[rtype][numGraph]
        //The origin is vconn[comp][0]
        vDesc_u dd=graphs[rtype][numGraph]->findVertex(gk[rtype].getId(v)).first;
        graphs[rtype][numGraph]->insertEdge(vconn[comp][0],dd,c.cost);


        //Insert path in solution
        vector<int> edges;
        for(list<vDesc_d>::iterator it=c.l.begin();it!=c.l.end();++it) {
            int id=gk[rtype].getId(*it);
            edges.push_back(id);
            sol.insert(id);
        }




        //set to 0 the cost of the edges in all gk[k]. We must set to 0 the cost of all the arcs
        //incident to the nodes in the path. Since we always have arcs (u,v) and (v,u) we search
        //the adjacent nodes for node u ande set to 0 all the arcs (v,u)
        for(int i=0;i<(int)edges.size();i++) {
            for(int k=0;k<K;k++) {
                pair<vDesc_d,bool> pv=gk[k].findVertex(edges[i]);
                if(!pv.second) continue;
                vDesc_d u=pv.first;
                aIter_d ai,aend;
                for(tie(ai,aend)=gk[k].getAdjVertices(u);ai!=aend;++ai) {
                    pair<eDesc_d,bool> pe=gk[k].getEdge(*ai,u);
                    gk[k].setEdgeCost(pe.first,0);
                }

            }
        }

        //The number of components in graphs[rtype][numGraphs] has decreased by 1.
        //If it was 2 then numGraph can't be choosen again
        if(n==2) { //now it is 1
            vector<int>::iterator it=lower_bound(availableGraphs[rtype].begin(),
                                                 availableGraphs[rtype].end(),numGraph);

            availableGraphs[rtype].erase(it);
            //Se não existem mais então tipo deixa de poder ser sorteado
            if(!availableGraphs[rtype].size()) {
                it=lower_bound(availableTypes.begin(),availableTypes.end(),rtype);
                availableTypes.erase(it);
            }
        }

    }

    for(int i=0;i<K;i++) {
        for(size_t j=0;j<graphs[i].size();j++) delete graphs[i][j];
    }
    delete [] graphs;

}


void CThreadsGreedy::makeMinimal(set<int>& sol)
{

    CGraph_u *ngk=createGraphsFromSolution(sol);

    //Try to remove vertices from graphs


    for(set<int>::iterator it=sol.begin();it!=sol.end();++it) {
        //Do not try to remove terminals
        bool isTerm=false;
        for(int k=0;k<K;k++) {
            if(binary_search(terminals[k].begin(),terminals[k].end(),*it)) {isTerm=true; break;}
        }
        if(isTerm) continue;

        //bool remover=true;
        vector<vDesc_u> removidos_adj[K];
        vector<int> retirados;
        for(int k=0;k<K;k++) {
            //If vertex is not in ngk continue

            vDesc_u idx_gk;
            bool flag;
            tie(idx_gk,flag)=ngk[k].findVertex((size_t) *it);
            if(!flag) continue;

            if(!removeVertex(ngk[k],idx_gk,k,removidos_adj[k])) {
                //remover=false;
                //put it back in ngk[k]
                vDesc_u v=ngk[k].insertVertex((size_t)*it);
                for(size_t ii=0;ii<removidos_adj[k].size();ii++) {
                    ngk[k].insertEdge(v,removidos_adj[k][ii],0);
                }
                //pit back the vertices removed earlier
                for(size_t j=0;j<retirados.size();j++) {
                    //Vertices
                    vDesc_u v=ngk[retirados[j]].insertVertex((size_t) *it);
                    //Adjacencies
                    for(size_t ii=0;ii<removidos_adj[retirados[j]].size();ii++) {
                        ngk[retirados[j]].insertEdge(v,removidos_adj[retirados[j]][ii],0);
                    }
                }
                break;
            }
            else retirados.push_back(k);
        }
    }

    //rewrite solution

    sol.clear();
    vIter_u vi,vi_end;
    for(int k=0;k<K;k++) {
        for(tie(vi,vi_end)=ngk[k].getVertices();vi!=vi_end;++vi)
            sol.insert(ngk[k].getId(*vi));
    }

    delete [] ngk;

}

CGraph_u *CThreadsGreedy::createGraphsFromSolution(set<int>&sol)
{
    CGraph_u *ngk;
    ngk=new CGraph_u [K];

    //Insert vertices from solution
    for(set<int>::iterator it=sol.begin();it!=sol.end();++it) {
        for(int k=0;k<K;k++) {
            if(gk[k].findVertex(*it).second) {
                ngk[k].insertVertex(*it);
            }
        }
    }

    //Insert edges
    for(int k=0;k<K;k++) {
        eIter_d ei,ei_end;
        bool flag;
        vDesc_u ori,dest;
        for(tie(ei,ei_end)=gk[k].getEdges();ei!=ei_end;++ei) {
            vDesc_d u=gk[k].getEdgeSource(*ei);
            tie(ori,flag)=ngk[k].findVertex(gk[k].getId(u));
            if(!flag) continue;
            vDesc_d v=gk[k].getEdgeTarget(*ei);
            tie(dest,flag)=ngk[k].findVertex(gk[k].getId(v));
            if(!flag) continue;
            if(!ngk[k].getEdge(ori,dest).second)
                ngk[k].insertEdge(ori,dest,0);
        }
    }

    for(int k=0;k<K;k++) {
        vector< vector<vDesc_u> > vconn;
        ngk[k].connComp(vconn);
        for(size_t i=0;i<vconn.size();i++) {
            bool hasTerm=false;
            for(size_t j=0;j<vconn[i].size();j++) {
                int id=ngk[k].getId(vconn[i][j]);
                if(binary_search(terminals[k].begin(),terminals[k].end(),id)) {hasTerm=true; break;}
            }
            if(!hasTerm) { //Remove component from graph
                for(size_t j=0;j<vconn[i].size();j++) {
                    ngk[k].removeVertex(vconn[i][j]);
                }
            }
        }
    }
    return ngk;
}


bool CThreadsGreedy::removeVertex(CGraph_u& ngk,vDesc_u idx,int k,vector<vDesc_u>& adjs)
{
    //idx is a descriptor if gk
    //before removing the vertex save the adjacencies

    aIter_u ai,ai_end;
    tie(ai,ai_end)=ngk.getAdjVertices(idx);
    while(ai!=ai_end) {
        adjs.push_back(*ai);
        ai++;
    }

    ngk.removeVertex(idx);

    //Calculate connected components

    vector< vector<vDesc_u> > vconn;
    ngk.connComp(vconn);
    //vconn contains descriptors of ngk.

    vector< set<int> > comps;
    for(size_t i=0;i<vconn.size();i++) {
        set<int> x;
        comps.push_back(x);
        for(size_t j=0;j<vconn[i].size();j++) {
            comps[i].insert(ngk.getId(vconn[i][j])); //change to ids
        }
    }

    //If any component is violated we can't remove
    for(size_t i=0;i<comps.size();i++) {
        if(fviol(comps[i],k)) return false;
    }

    return true;
}

bool CThreadsGreedy::fviol(const set<int>& conn,int k)
{
    //Check if any component is violated. A component is violated if for
    //any pair of terminals it contains one but not the other
    //reurns: true if the component is violated, false otherwise

    bool existe1,existe2;
    for(set< pair<int,int> >::iterator it=term_par[k].begin();it!=term_par[k].end();it++) {
        existe1=false; existe2=false;
        if(conn.find(it->first)!=conn.end()) existe1=true;
        if(conn.find(it->second)!=conn.end()) existe2=true;
        if(existe1!=existe2) return true;
    }

    return false;
}
void CThreadsGreedy::calcSolutionByType(const set<int>& sol)
{
    for(int k=0;k<K;k++) {
        CGraph_d *rg=new CGraph_d;


        //Insert vertices
        for(set<int>::iterator it=sol.begin();it!=sol.end();++it) {
            if(gk[k].findVertex(*it).second) rg->insertVertex(*it);
        }

        //Insert the arcs.
        eIter_d ei,ei_end;
        for(tie(ei,ei_end)=gk[k].getEdges();ei!=ei_end;++ei) {
            //Start by obtaining the Ids in gk
            size_t id_u,id_v;
            id_u=gk[k].getId(gk[k].getEdgeSource(*ei) );
            id_v=gk[k].getId(gk[k].getEdgeTarget(*ei) );
            vDesc_d idx_u_rg,idx_v_rg;
            bool flag;
            tie(idx_u_rg,flag)=rg->findVertex(id_u);
            if(!flag) continue;
            tie(idx_v_rg,flag)=rg->findVertex(id_v);
            if(!flag) continue;
            //Insert arc u-->v
            //Calculate the cost
            double c=data->getCost(id_v);
            rg->insertEdge(idx_u_rg,idx_v_rg,c);
        }

        //rg contains ids of sites and terminals

        //To find the solution for each type
        //calculate the shortest path between all pairs of reserves of this type
        //We have to calculate paths starting at 0,...,terminals[i].size()-2 to all the others

        set<size_t> thisSol;
        for(int j=0;j<(int)terminals[k].size()-1;++j) {

            bool exists;
            vDesc_d ori;
            size_t id=terminals[k][j];
            tie(ori,exists)=rg->findVertex(id);
            if(!exists)continue;

            vector<vDesc_d> dest;
            for(size_t jj=j+1;jj<terminals[k].size();++jj) {
                vDesc_d v;
                id=terminals[k][jj];
                tie(v,exists)=rg->findVertex(id);
                if(!exists) continue;
                dest.push_back(v);
            }
            if(!dest.size()) continue;
            vector<caminho_d_t> path;
            rg->dijkstra(ori,dest,path);

            //Save solution
            for(size_t jj=0;jj<path.size();++jj) {
                if(!path[jj].l.size()) continue;
                //convert descriptots in the path to ids
                for(list<vDesc_d>::iterator it=path[jj].l.begin();it!=path[jj].l.end();++it) {
                    thisSol.insert(rg->getId(*it));
                }
            }
        }

        delete rg;

        //Calculate cost and save solution (terminals are ignored)
        double c=0;
        for(set<size_t>::iterator it=thisSol.begin();it!=thisSol.end();++it) {
            const vertex_data *vd=data->getVertexDataById(*it);
            if(vd->siteType==K+1) { //Multiple types
                if(!binary_search(vd->cellTypes.begin(),vd->cellTypes.end(),k+1)) {
                    c+=data->getCost(*it);
                    resSol->sol[k].vertices.push_back(*it);
                }
            }
            else if(vd->siteType!=(int)k+1) {
                c+=data->getCost(*it);
                resSol->sol[k].vertices.push_back(*it);
            }
        }
        resSol->sol[k].cost=c;

    }
}
